"""Controller for Port object. Mostly handles conversion between mongo data and python objects"""

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.controllers.controllerelement import ControllerElement
import json

class PortController(ControllerElement):
    """Inherits ControllerElement
    Controller for Port object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Port represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by PortView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Port document.
        """
        self.model.service = values.get("Service", self.model.service)
        self.model.product = values.get("Product", self.model.product)
        self.model.notes = values.get("Notes", self.model.notes)
        self.model.infos = json.loads(values.get("Infos", self.model.infos))
        self.model.update()

    def doInsert(self, values):
        """
        Insert the Port represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by PortView containg all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        # get form values
        port = values["Number"]
        proto = values["Proto"]
        service = values["Service"]
        product = values["Product"]
        notes = values["Product"]
        # Add the port in database
        self.model.initialize(values["ip"], port, proto, service, product, notes=notes)
        ret = self.model.addInDb()
        return ret, 0  # 0 errors

    
    def addCustomTool(self, command_iid):
        """Add command iid to instantiate as a tool in the model 
        Args:
            command_iid:  command iid of the command to instantiate as a tool in the model 
        """
        self.model.addCustomTool(command_iid)

    def getDefects(self):
        """Return port assigned defects as a list of mongo fetched defects dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return self.model.getDefects()

    def getChecks(self):
        """Return ports assigned checks as a list of mongo fetched checks
        Returns:
            list of checkInstance objects
        """
        return self.model.getChecks()
    
    @classmethod
    def getDefectsForPorts(self, ports):
        datamanager = DataManager.getInstance()
        return datamanager.find("defect", {"target_iid": {"$in": [str(port.getId()) for port in ports]}})

    @classmethod
    def getChecksForPorts(self, ports):
        datamanager = DataManager.getInstance()
        return datamanager.find("checkinstance", {"target_iid": {"$in": [str(port.getId()) for port in ports]}})
    
    def getType(self):
        """Returns a string describing the type of object
        Returns:
            "port" """
        return "port"


import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import os
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.components.apiclient import APIClient
from PIL import Image, ImageTk
from ttkwidgets import CheckboxTreeview
import re
import importlib

class ScriptManager:
    """
    Open the scripts window manager
    """
    folder_icon = None

    def __init__(self) -> None:
        pass

    def initUI(self, appli, whatfor="execute"):
        """
            appli: the tkinter parent view to use for this window construction.
        """
        self.appli = appli
        parent = appli
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Scripts Manager")
        self.app.resizable(True, True)
        self.rvalue = None
        appFrame = CTkFrame(self.app)
         #PANED PART
        self.paned = tk.PanedWindow(appFrame, height=300)
        #RIGHT PANE : TAble
        self.viewframe = CTkFrame(self.paned)
        self.file_tree = CheckboxTreeview(self.viewframe)
        self.file_tree['columns'] = ('name', 'category')
        self.file_tree.heading('#0', text='Name')
        self.file_tree.column("#0", stretch=tk.YES, minwidth=300, width=300)
        self.file_tree.heading('#1', text='Category')
        self.file_tree.column("#1", stretch=tk.YES, minwidth=300, width=300)
        self.file_tree.pack(fill=tk.BOTH, expand=True)
        btn_pane = CTkFrame(self.viewframe)
        if "execute" in whatfor:
            self.execute_icone = CTkImage(Image.open(utils.getIcon("execute.png")))
            btn_execute = CTkButton(btn_pane, text="Execute", image=self.execute_icone, command=self.executedSelectedScripts)        
            btn_execute.pack(side=tk.RIGHT, padx=3, pady=5)
        elif "select" in whatfor:
            btn_select = CTkButton(btn_pane, text="Select", command=self.returnSelection)        
            btn_select.pack(side=tk.RIGHT, padx=3, pady=5)
        self.open_folder_icone = CTkImage(Image.open(utils.getIcon("folder.png")))
        btn_openPathForUser = CTkButton(btn_pane, text="Open folder", image=self.open_folder_icone, command=self.openPathForUser)        
        btn_openPathForUser.pack(side=tk.RIGHT, padx=3, pady=5)
        
        btn_pane.pack(fill=tk.X, side=tk.BOTTOM, anchor=tk.E)
        #LEFT PANE : Treeview
        self.frameTw = CTkFrame(self.paned)
        self.treevw = ttk.Treeview(self.frameTw)
        self.treevw.pack()
        self.rvalue = None
        scbVSel = CTkScrollbar(self.frameTw,
                                orientation=tk.VERTICAL,
                                command=self.treevw.yview)
        self.treevw.configure(yscrollcommand=scbVSel.set)
        self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
        scbVSel.grid(row=0, column=1, sticky=tk.NS)
        self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
        scbVSel.grid(row=0, column=1, sticky=tk.NS)
        self.paned.add(self.frameTw)
        self.paned.add(self.viewframe)
        self.paned.pack(fill=tk.BOTH, expand=1)
        self.frameTw.rowconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.frameTw.columnconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize

        appFrame.pack(fill=tk.BOTH, ipady=10, ipadx=10, expand=True)

        self.treevw.bind("<<TreeviewSelect>>", self.onTreeviewSelect)
        try:
            self.app.wait_visibility()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass
        self.refreshUI()

    def refreshUI(self):
        for widget in self.treevw.winfo_children():
            widget.destroy()
        script_dir = ScriptManager.getScriptsDir()
        if self.__class__.folder_icon is None:
            self.__class__.folder_icon = ImageTk.PhotoImage(file=utils.getIcon("folder.png"))
        parent = self.treevw.insert("", "end", " ", text="Scripts", image=self.__class__.folder_icon, open=True)
        self.treevw.focus(parent)
        self.treevw.selection_set(parent)
        for root, subFolders, files in os.walk(script_dir):
            root_name = root.replace(script_dir, "")
            for folder in subFolders:
                if folder.startswith("__") or folder.endswith("__"):
                    continue
                folder_iid = os.path.join(root_name, folder)
                parent_node = parent if root_name == "" else root_name
                self.treevw.insert(parent_node, "end", folder_iid, text=folder, image=self.__class__.folder_icon)
        self.openScriptFolderView()

    @staticmethod
    def getScriptsList(searchreq=""):
        liste = []
        script_dir = ScriptManager.getScriptsDir()
        for root, subFolders, files in os.walk(script_dir):
            root_name = root.replace(script_dir, "")
            for folder in subFolders:
                if folder.startswith("__") or folder.endswith("__"):
                    continue
                folder_iid = os.path.join(root_name, folder)
                if searchreq in folder_iid:
                    liste.append(folder_iid)
        return liste

    @staticmethod
    def getScriptsDir():
        return os.path.normpath(os.path.join(utils.getMainDir(), "scripts/"))

    def onTreeviewSelect(self, _event=None):
        
        selec = self.treevw.selection()
        if len(selec) == 0:
            return None
        item = selec[0]
        self.openScriptFolderView(str(item))

    def openScriptFolderView(self, script_folder=""):
        full_script_path = os.path.join(ScriptManager.getScriptsDir(), script_folder.strip())
        script_shown = set()
        self.file_tree.delete(*self.file_tree.get_children())

        for root, _, files in os.walk(full_script_path):
            for file in files:
                filepath = root + '/' + file
                if file.endswith(".py"):
                    script_shown.add(filepath)
        scripts_list = sorted(script_shown)
        script_dir = ScriptManager.getScriptsDir()
        for script in scripts_list:
            scriptName = os.path.basename(script)
            category_name = os.path.dirname(script.replace(script_dir, ""))
            self.file_tree.insert("", "end", os.path.normpath(script), text=scriptName, values=(category_name))
        

    def executedSelectedScripts(self):
        for selected in self.file_tree.get_checked():
            self.executeScript(self.appli, selected, parent=self.app)

    def returnSelection(self):
        ret = []
        folder = ScriptManager.getScriptsDir()
        for selected in self.file_tree.get_checked():
            common = os.path.commonpath([folder, selected])
            ret_line = selected.replace(common, "")
            ret_line = re.sub(r"^/+", "", ret_line)
            ret.append(ret_line)

        self.rvalue = ret
        self.app.destroy()

    def executeScript(self, appli, script_path, data={}, parent=None):
        script_dir = ScriptManager.getScriptsDir()
        category_name = os.path.dirname(script_path.replace(script_dir, ""))
        category_name = re.sub(r"^/+", "",  category_name)
        script_name = ".".join(os.path.splitext(os.path.basename(script_path))[:-1])
        module = os.path.join("pollenisatorgui/scripts/",category_name, script_name).replace("/", '.')
        imported = importlib.import_module(module)
        success, res = imported.main(APIClient.getInstance(), appli, **data)
        if success:
            tk.messagebox.showinfo("Script finished", f"Script {script_name} finished.\n{res}", parent=parent)
        else:
            tk.messagebox.showwarning("Script failed", f"Script {script_name} failed.\n{res}", parent=parent)

    def openPathForUser(self):
        selection = self.treevw.selection()
        if selection:
            folder = os.path.join(ScriptManager.getScriptsDir(), selection[0])
        else:
            folder = ScriptManager.getScriptsDir()
        utils.openPathForUser(folder)

    @classmethod
    def openScriptForUser(cls, script):
        script = re.sub(r"^/+", "", script)
        folder = os.path.join(cls.getScriptsDir(), script)
        utils.openPathForUser(folder)


"""Controller for command object. Mostly handles conversion between mongo data and python objects"""
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.controllers.controllerelement import ControllerElement
import bson

class CheckInstanceController(ControllerElement):
    """Inherits ControllerElement
    Controller for CheckInstance object. Mostly handles conversion between mongo data and python objects"""
    def doUpdate(self, values):
        """
        Update the command represented by this self.model in database with the given values.

        Args:
            values: A dictionary crafted by CheckInstanceView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated command document.
        """
        # Get form values
        self.model.status = values.get(
            "Status", self.model.status)
        self.model.notes = values.get(
            "Notes", self.model.notes)
        
        # Update in database
        self.model.update()

    def getData(self):
        """Return command attributes as a dictionnary matching Mongo stored commands
        Returns:
            dict with keys name, lvl, safe, text, ports, priority, max_thread, priority, types, _id, tags and infos
        """
        return self.model.getData()

    def getCheckItem(self):
        return self.model.getCheckItem()

    def getCheckInstanceStatus(self):
        return self.model.getCheckInstanceStatus()

    def getCheckInstanceInfo(self):
        return APIClient.getInstance().getCheckInstanceInfo(self.model.getId())

    def getType(self):
        """Return a string describing the type of object
        Returns:
            "checkinstance" """
        return "checkinstance"

    def getStatus(self):
        return self.model.getStatus()
    
    def swapStatus(self):
        curr = self.getStatus()
        if curr == "done":
            self.model.status = "todo"
        elif curr == "todo" or curr == "running" or curr == "":
            self.model.status = "done"
        else:
            return
        self.model.update({"status": self.model.status})

    @property
    def target_repr(self):
        if hasattr(self.model, "target_repr"):
            return self.model.target_repr
        reprs = APIClient.getInstance().getCheckInstanceRepr([str(self.model.getId())])
        if reprs:
            self.model.target_repr = reprs.get(str(self.model.target_iid), str(self.model))
        else:
            self.model.target_repr = str(self)
        return self.model.target_repr
    
    @target_repr.setter
    def target_repr(self, value):
        self.model.target_repr = value

    def getCategory(self):
        return self.model.check_m.category
    
    def getTarget(self):
        return self.model.target_iid
    
    def getTools(self):
        return self.model.getTools()
        
    def isAuto(self):
        return self.model.check_m.isAuto()

    def actualize(self):
        """Ask the model to reload its data from database
        """
        if self.model is not None:
            self.model = self.model.__class__.fetchObject({"_id": bson.ObjectId(self.model.getId())})

   


"""CheckItem Model."""
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.models.tool import Tool
from bson import ObjectId

class CheckInstance(Element):
    """Represents a CheckInstance object of a cheatsheet.

    Attributes:
        coll_name: collection name in pollenisator or pentest database
    """

    coll_name = "checkinstances"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched command is optimal.
                        possible keys with default values are : _id (None), parent (None), check_iid, status, notes
        """
        if valuesFromDb is None:
            valuesFromDb = dict()
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None), valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("check_iid"), valuesFromDb.get("target_iid"),valuesFromDb.get("target_type"), valuesFromDb.get("parent"), valuesFromDb.get("status", ""), valuesFromDb.get("notes", ""))
        
   
    def initialize(self, check_iid, target_iid, target_type, parent, status, notes):
        self.check_iid = check_iid
        self.parent = parent
        self.target_iid = target_iid
        self.target_type = target_type
        self.status = status
        self.notes = notes
        from pollenisatorgui.core.components.datamanager import DataManager
        datamanager = DataManager.getInstance()
        self.check_m = datamanager.get("checkitems", str(check_iid))
        if self.check_m is None:
            self.check_m = CheckItem.fetchObject({"_id":ObjectId(check_iid)})
        if self.check_m is None:
            #print(" Warning : Check item not found. Might have been deleted")
            return None
        return self

    def delete(self):
        ret = self._id
        apiclient = APIClient.getInstance()
        apiclient.deleteCheckInstance(ret)
        

    # def addInDb(self):
    #     """Add this command to pollenisator database
    #     Returns: a tuple with :
    #             * bool for success
    #             * mongo ObjectId : already existing object if duplicate, create object id otherwise 
    #     """
    #     apiclient = APIClient.getInstance()
    #     res, id = apiclient.insertCheckInstance(self.getData())
    #     if not res:
    #         return False, id
    #     self._id = id
    #     return True, id
        
    def getStatus(self):
        return self.status
    

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            apiclient.updateCheckInstance(self._id, self.getData())
        else:
            apiclient.updateCheckInstance(self._id, pipeline_set )
        


    @classmethod
    def fetchObject(cls,  pipeline):
        """Fetch one CheckInstance from database and return the CheckInstance object 
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a CheckInstance or None if nothing matches the pipeline.
        """
        apiclient = APIClient.getInstance()
        res = apiclient.findCheckInstance(pipeline, many=False)
        d = res
        if d is None:
            return None
        return CheckInstance(d)

    @classmethod
    def fetchObjects(cls, pipeline):
        """Fetch many cheatsheet from database and return a Cursor to iterate over cheatsheet objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on cheatsheet objects
        """
        apiclient = APIClient.getInstance()
        if isinstance(pipeline, list):
            pipeline = {"_id": {"$in":list(map(ObjectId, pipeline))}}
        ds = apiclient.findCheckInstance(pipeline)
        if ds is None:
            return None
        for d in ds:
            yield CheckInstance(d)
    
    def getData(self):
        return {"_id": self._id,  "check_iid": self.check_iid, "target_type": self.target_type, "target_iid": self.target_iid,"parent":self.parent, "status": self.status, "notes": self.notes}

    def getCheckItem(self):
        return self.check_m

    def getTools(self):
        datamanager = DataManager.getInstance()
        return datamanager.find("tools", {"check_iid": str(self._id)})

    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (1 key :"name")
        """
        return {"check_iid": self.check_iid}

    def __str__(self):
        if self.check_m is None:
            return "ERROR"
        return self.check_m.title

    def getCheckInstanceStatus(self):
        """
        Get a more detailed string representation of a check instance.
        Returns:
            string
        """
        data = self.getData()
        if self.check_m is None:
            check_item_data = {}
        else:
            check_item_data = self.check_m.getData()
        data["check_item"] = check_item_data
        data["tools_done"] = {}
        data["tools_running"] = {}
        data["tools_not_done"] = {}
        all_complete = True
        at_least_one = False
        total = 0
        done = 0
        tools_to_add = self.getTools()
        if tools_to_add is not None:
            for tool in tools_to_add:
                if "done" in tool.getStatus():
                    done += 1
                    at_least_one = True
                    data["tools_done"][str(tool.getId())] = tool.getData()
                elif "running" in tool.getStatus():
                    at_least_one = True
                    data["tools_running"][str(
                        tool.getId())] = tool.getData()
                else:
                    data["tools_not_done"][str(
                        tool.getId())] = tool.getData()
                total += 1

        if done != total:
            all_complete = False
        if self.check_m is not None:
            if len(self.check_m.commands) > 0:
                if at_least_one and all_complete:
                    data["status"] = "done"
                elif at_least_one and not all_complete:
                    data["status"] = "running"
                else:
                    data["status"] = "todo"
            else:
                data["status"] = ""
        else:
            data["status"] = "error"
        return data



"""View for scope object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.controllers.checkinstancecontroller import CheckInstanceController
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
from pollenisatorgui.core.views.ipview import IpView
from pollenisatorgui.core.views.multiplescopeview import MultipleScopeView
from pollenisatorgui.core.controllers.scopecontroller import ScopeController
from pollenisatorgui.core.models.scope import Scope
from pollenisatorgui.core.views.viewelement import ViewElement
from tkinter import TclError

class ScopeView(ViewElement):
    """View for port object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory"""

    icon = 'scope.png'

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Scope
        """
        self.form.clear()
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Scope", column=0)
        top_panel.addFormStr("Scope", '', modelData["scope"], None, column=1, state="readonly")
        notes = modelData.get("notes", "")
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Notes")
        self.form.addFormText("Notes", r"", notes, None, side="top")
        self.completeModifyWindow()

    
    def addChildrenBaseNodes(self, newNode):
        """
        Add to the given node from a treeview the mandatory childrens.
        For a Scope it is the tools parent node and the ips parent node

        Args:
            newNode: the newly created node we want to add children to.
        """

        return self.appliTw.insert(newNode, "end", IpView.DbToTreeviewListId(newNode), text="IPs", image=IpView.getClassIcon())
    
    def openInsertWindow(self):
        view = MultipleScopeView(self.appliTw, self.appliViewFrame, self.mainApp, self.controller)
        view.openInsertWindow()
    
    def getAdditionalContextualCommands(self):
        return {"Insert Scopes":self.openInsertWindow}

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            addChildren: If False, skip the tool insert. Useful when displaying search results
        """
        addChildren = kwargs.get("addChildren", True)
        self.appliTw.views[str(self.controller.getDbId())] = {"view":self}

        parentDbId = parentNode
        if parentNode is None:
            parentNode = self.getParentNode()
        elif parentNode == "":
            parentNode = ""
        elif 'scopes' not in parentNode:
            parentNode = ScopeView.DbToTreeviewListId(parentDbId)
        if parentNode != "" and parentNode is not None and kwargs.get("insert_parents", True):
            try:
                parentNode = self.appliTw.insert(
                    self.controller.getParentId(), 0, parentNode, text="Scopes", image=self.getClassIcon())
            except TclError as e:  #  trigger if tools list node already exist
                pass
        try:
            node = self.appliTw.insert(parentNode, "end", str(self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=self.controller.getTags(), image=self.getClassIcon())
        except TclError as e:
            pass
        if addChildren:
            self._insertChildren()
        elif self.appliTw.lazyload and not self.mainApp.searchMode:
            try:
                self.appliTw.insert(self.controller.getDbId(), "end", str(self.controller.getDbId()+"|<Empty>"), text="<Empty>")
            except TclError as e:
                pass
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")
        if "hidden" in self.controller.getTags():
            self.hide("tags")

    def _insertChildren(self):
        checks = self.controller.getChecks()
        for check in checks:
            check_o = CheckInstanceController(check)
            check_vw = CheckInstanceView(self.appliTw, self.appliViewFrame, self.mainApp, check_o)
            check_vw.addInTreeview(str(self.controller.getDbId()))

    @classmethod
    def treeviewListIdToDb(cls, treeview_id):
        """Extract from the unique string identifying a list of scopes the parent db ID
        Args:
            treeview_id: the treeview node id of a list of scopes node
        Returns:
            the parent object mongo id as string
        """
        return str(treeview_id).split("|")[1]

    @classmethod
    def DbToTreeviewListId(cls, parent_db_id):
        """Converts a mongo Id to a unique string identifying a list of scopes given its parent
        Args:
            parent_db_id: the parent node mongo ID
        Returns:
            A string that should be unique to describe the parent list of scope node
        """
        return "scopes|"+str(parent_db_id)

    def split_ip(self):
        """Split a IP address given as string into a 5-tuple of integers.
        Returns:
            If network IP Tuple of 5 integers values representing the 4 parts of an ipv4 string + the /mask integer
            Otherwise returns self"""
        modelData = self.controller.getData()
        try:
            ret = tuple(int(part) for part in modelData["scope"].split('.'))
            ret = ret + tuple(int(modelData["scope"].split('/')[1]))
        except ValueError:
            ret = tuple(str(part) for part in modelData["scope"].split('.'))
        return ret

    def insertReceived(self):
        """Called when a scope insertion is received by notification.
        Tells the parent wave to update itself
        """
        if self.controller.model is None:
            return
        parentId = self.controller.getParentId()
        parentView = self.appliTw.views[str(parentId)]["view"]
        parentView.updateReceived(None, None)
        super().insertReceived()

    def key(self):
        """Returns a key for sorting this node
        Returns:
            Tuple of 5 integer valus representing the scope perimeter if network ip or self directly
        """
        return self.split_ip()



import logging
import sys
import os
formatter = logging.Formatter("%(asctime)s;%(levelname)s;%(message)s",
                              "%Y-%m-%d %H:%M:%S")
log_path = os.path.expanduser("~/.pollenisator")
os.makedirs(log_path, exist_ok=True)
logging.basicConfig(filename=os.path.join(log_path, 'debug.log'), encoding='utf-8', level=logging.INFO)
logger = logging.getLogger(__name__)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)

logger.addHandler(handler)

def handle_exception(exc_type, exc_value, exc_traceback):
    logger.error("Exception", exc_info=(exc_type, exc_value, exc_traceback))
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

sys.excepthook = handle_exception


"""Ttk treeview class with added functions.
"""
import tkinter as tk
from bson.objectid import ObjectId
from pollenisatorgui.core.application.treeviews.PollenisatorTreeview import PollenisatorTreeview
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion

from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.core.views.checkitemview import CheckItemView
from pollenisatorgui.core.views.multicheckitemView import MultiCheckItemView
from pollenisatorgui.core.controllers.checkitemcontroller import CheckItemController
import pollenisatorgui.core.components.utils as utils


class CheatsheetTreeview(PollenisatorTreeview):
    """CheatsheetTreeview class
    Inherit PollenisatorTreeview.
    Ttk treeview class with added functions to handle the command objects.
    """

    def __init__(self, appli, parentFrame, viewFrame):
        """
        Args:
            appli: a reference to the main Application object.
            parentFrame: the parent tkinter window object.
        """
        super().__init__(appli, parentFrame)
        self.viewFrame = viewFrame
        self.openedViewFrameId = None  # if of the currently opened object in the view frame

    def attachViewFrame(self, viewFrame):
        self.viewFrame = viewFrame


    def initUI(self, _event=None):
        """Initialize the user interface widgets and binds them.
        Args:
            _event: not used but mandatory
        """
        self._initContextualsMenus()
        self.title = "Cheatsheet"
        self.heading('#0', text='Checks', anchor=tk.W)
        self.column('#0', stretch=tk.YES, minwidth=300, width=300)
        self.bind("<Button-3>", self.doPopup)
        self.bind("<<TreeviewSelect>>", self.onTreeviewSelect)
        #self.bind("<Return>", self.onTreeviewSelect)
        #self.bind("<Button-1>", self.onTreeviewSelect)
        self.bind('<Delete>', self.deleteSelected)
    
    def onTreeviewSelect(self, event=None):
        """Called when a line is selected on the treeview
        Open the selected object view on the view frame.
        IF it's a parent commands or other node, opens Insert
        ELSE open a modify window
        Args:
            event: filled with the callback, contains data about line clicked
        """
        selection = self.selection()
        if len(selection) == 1:
            item = super().onTreeviewSelect(event)
            if isinstance(item, str):
                pass
                # if "pentest|" not in str(item):
                #     objView = CheckItemView(
                #         self, self.viewFrame, self.appli, CheckItemController(CheckItem({"category":str(item)})))
                #     objView.openInsertWindow()
            else:
                self.openModifyWindowOf(item)
        elif len(selection) > 1:
            # Multi select:
            multiView = MultiCheckItemView(self, self.viewFrame, self.appli)
            for widget in self.viewFrame.winfo_children():
                widget.destroy()
            multiView.form.clear()
            multiView.openModifyWindow()

    def add_check(self, _event=None):
        item = self.contextualMenu.selection
        if item is not None:
            objView = self.getViewFromId(str(item))
            if objView is None:
                if "pentest|" not in str(item):
                    objView = CheckItemView(
                        self, self.viewFrame, self.appli, CheckItemController(CheckItem({"category":str(item)})))
                    objView.openInsertWindow()
            else:
                new = CheckItemView(
                        self, self.viewFrame, self.appli, CheckItemController(CheckItem({"category":str(objView.controller.getCategory())})))
                new.openInsertWindow()

    def openInsertWindow(self, check_item):
        objView = CheckItemView(
                        self, self.viewFrame, self.appli, CheckItemController(check_item))
        objView.openInsertWindow()

    def doPopup(self, event):
        """Open the popup 
        Args:
            event: filled with the callback, contains data about line clicked
        """
        self.contextualMenu.selection = self.identify(
            "item", event.x, event.y)
        super().doPopup(event)

    def openModifyWindowOf(self, dbId):
        """
        Retrieve the View of the database id given and open the modifying form for its model and open it.

        Args:
            dbId: the database Mongo Id to modify.
        """
        objView = self.getViewFromId(str(dbId))
        if objView is not None:
            for widget in self.viewFrame.winfo_children():
                widget.destroy()
            objView.form.clear()
            self.openedViewFrameId = str(dbId)
            objView.openModifyWindow(treevw="cheatsheet")

    def load(self, _searchModel=None):
        """
        Load the treeview with database information

        Args:
            _searchModel: (Deprecated) inherited not used. 
        """
        for widget in self.viewFrame.winfo_children():
            widget.destroy()
        self.delete(*self.get_children())

        self._load()

    def _load(self):
        """
        Load the treeview with database information
        """
       
        checkitems = CheckItem.fetchObjects({})
        checkitems = sorted(checkitems, key=lambda x:x.step)
        for checkitem in checkitems:
            checkitem_vw = CheckItemView(
                self, self.viewFrame, self.appli, CheckItemController(checkitem))
            checkitem_vw.addInTreeview(with_category=True)

            
    def deleteSelected(self, _event=None):
        """
        Interface to delete a database object from an event.
        Prompt the user a confirmation window.
        Args:
            _event: not used, a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        n = len(self.selection())
        dialog = ChildDialogQuestion(self.parentFrame,
                                     "DELETE WARNING", "Becareful for you are about to delete "+str(n) + " entries and there is no turning back.", ["Delete", "Cancel"])
        self.wait_window(dialog.app)
        if dialog.rvalue != "Delete":
            return
        if n == 1:
            view = self.getViewFromId(self.selection()[0])
            if view is None:
                return
            view.delete(None, False)
        else:
            toDelete = {}
            for selected in self.selection():
                view = self.getViewFromId(selected)
                if view is not None:
                    viewtype = view.controller.model.coll_name
                    if viewtype not in toDelete:
                        toDelete[viewtype] = []
                    toDelete[viewtype].append(view.controller.getDbId())
            apiclient = APIClient.getInstance()
            apiclient.bulkDelete(toDelete)

    def refresh(self, force=True):
        """Alias to self.load method"""
        self.load()

    def applyToPentest(self, node=None):
        """Add the current check to the pentest check instances where it applies
        """
        if node is None:
            try:
                node = str(self.contextualMenu.selection)
            except:
                return
        apiclient = APIClient.getInstance()
        apiclient.apply_check_to_pentest(str(node))

    def _initContextualsMenus(self):
        """
        Create the contextual menu
        """
        self.contextualMenu = utils.craftMenuWithStyle(self.parentFrame)
        self.contextualMenu.add_command(label="Add new Check", command=self.add_check)
        self.contextualMenu.add_command(
            label="Apply to opened pentest", command=self.applyToPentest)
        self.contextualMenu.add_separator()
        self.contextualMenu.add_command(
            label="Sort children", command=self.sort)
        self.contextualMenu.add_command(
            label="Expand", command=self.expand)
        self.contextualMenu.add_command(
            label="Collapse", command=self.collapse)
        self.contextualMenu.add_separator()
        self.contextualMenu.add_command(
            label="Close", command=self.closeMenu)
        super()._initContextualsMenus
        return self.contextualMenu

    def update_received(self, dataManager, notif, obj, old_obj):
        collection = notif["collection"]
        action = notif["action"]
        iid = notif["iid"]
        if collection != "checkitems":
            return
        # Delete
        apiclient = APIClient.getInstance()
        if action == "delete":
            try:
                self.delete(ObjectId(iid))
            except tk.TclError:
                pass  # item was not inserted in the treeview

        # Insert
        if action == "insert":
            if collection == "checkitems":
                checkitem = CheckItem.fetchObject({"_id":ObjectId(iid)})
               
                if checkitem is not None:
                    view = CheckItemView(self, self.viewFrame,
                                   self.appli, CheckItemController(checkitem))
                    parent = None
                    try:
                        view.addInTreeview(parent, addChildren=True)
                        if view is not None:
                            view.insertReceived()
                    except tk.TclError:
                        pass

        if action == "update":
            try:
                view = self.getViewFromId(str(iid))
                if view is not None:
                    self.item(str(iid), text=str(
                        view.controller.getModelRepr()), image=view.getIcon())
            except tk.TclError:
                if view is not None:
                    view.addInTreeview()
            if str(iid) == str(self.openedViewFrameId):
                self.after(105, view.reopenView)
                
            if view is not None:
                view.controller.actualize()
                view.updateReceived()



"""View for command list object. Present an multi editing form to userh."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.viewelement import ViewElement
import pollenisatorgui.core.components.utils as utils

class MultiCommandView(ViewElement):
    """View for command list object. Present an multi editing form to user when interacted with."""
    
    def __init__(self, appliTw, appViewFrame, mainApp):
        super().__init__(appliTw, appViewFrame, mainApp, None)

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to manipulate many commands
        """
        self.form.clear()
        top_panel = self.form.addFormPanel()
        top_panel.addFormButton("Add to my commands", self.addSelectedToMyCommands)
        top_panel.addFormButton("Add to Worker commands", self.addSelectedToWorkerCommands)
        top_panel.addFormButton("Remove selection from my commands", self.removeSelectedFromMyCommands)
        top_panel.addFormButton("Delete", self.appliTw.deleteSelected, image=self.delete_image,
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        self.showForm()

    def addSelectedToMyCommands(self, event=None):
        apiclient = APIClient.getInstance()
        for selected in self.appliTw.selection():
            apiclient.addCommandToMyCommands(selected)

    def addSelectedToWorkerCommands(self, event=None):
        apiclient = APIClient.getInstance()
        for selected in self.appliTw.selection():
            apiclient.addCommandToWorkerCommands(selected)

    def removeSelectedFromMyCommands(self, event=None):
        apiclient = APIClient.getInstance()
        for selected in self.appliTw.selection():
            apiclient.removeCommandFromMyCommands(selected)


"""Controller for model object. Mostly handles conversion between mongo data and python objects"""

from bson.objectid import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.tags import Tag


class ControllerElement:
    """Controller for model object. Mostly handles conversion between mongo data and python objects"""
    def __init__(self, model):
        """Constructor
        Args:
            model: Any instance of classe that inherits core.Models.Element"""
        self.model = model

    def getDbId(self):
        """Returns the mongo database id of the model object
        Returns:
            bson.objectid ObjectId or None if self.model is None
        """
        if self.model is not None:
            return self.model.getId()
        return None

    def getParentId(self):
        """Return the parent object database id of the model. E.G a port would returns its parent IP mongo id
        Returns:
            None if model is None
            bson.objectid ObjectId of the parent object
        """
        if self.model is None:
            return None
        return self.model.getParentId()

    def getData(self):
        if self.model is None:
            return {}
        return self.model.getData()

    def doDelete(self):
        """Ask the model to delete itself from database
        """
        self.model.delete()

    def actualize(self):
        """Ask the model to reload its data from database
        """
        if self.model is not None:
            self.model = self.model.__class__.fetchObject(
                {"_id": ObjectId(self.model.getId())})

    def update(self):
        """Update object in database with model data
        """
        self.model.update()

    def updateInfos(self, infos):
        """Update object in database with given dictionnary
        Args:
            infos: a dictionnary with updated values for this object.
        """
        self.model.updateInfos(infos)

    def getModelRepr(self, detailed=False):
        """Returns a string representation of the model
        Returns:
            a string conversion of the model
        """
        try:
            if detailed:
                return self.model.getDetailedString()
            return str(self.model)
        except TypeError:
            return "Error"

    def getTags(self):
        """Returns a list of string decribing tags
        Returns:
            list of string
        """
        if self.model is None:
            return
        datamanager = DataManager.getInstance()
        res = datamanager.find("tags", {"item_id":ObjectId(self.model.getId())}, multi=False, fetch_on_none=False)
        if res is None:
            return []
        return res.tags

    def addTag(self, newTag, overrideGroupe=True):
        """Add the given tag to this object.
        Args:
            newTag: a new tag as a string to be added to this model tags
            overrideGroupe: Default to True. If newTag is in a group with a tag already assigned to this object, it will replace this old tag.
        """
        Tag.addTagTo(self.getDbId(), self.getType(), newTag, overrideGroupe)

    def delTag(self, tagToDelete):
        """Delete the given tag in this object.
        Args:
            tagToDelete: a tag as a string to be deleted from this model tags
        """
        Tag.delTag(self.getDbId(), self.getType(), tagToDelete)

    def setTags(self, tags):
        """Change all tags for the given new ones and update database
        Args:
            tags: a list of tag string
        """
        Tag.setTags(self.getDbId(), self.getType(), tags)

    def getDetailedString(self):
        """Return a string describing the model with more info than getModelRepr. E.G a port goes from "tcp/80" to "IP.IP.IP.IP tcp/80"
        Returns:
            a detailed string conversion of the model
        """
        return self.model.getDetailedString()

    def getType(self):
        """Return a string describing the type of object
        Returns:
            "element" """
        return "element"


"""Controller for Scope object. Mostly handles conversion between mongo data and python objects"""

import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.controllers.controllerelement import ControllerElement
from pollenisatorgui.core.models.scope import Scope


class ScopeController(ControllerElement):
    """Inherits ControllerElement
    Controller for Scope object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Scope represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by ScopeView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Scope document.
        """
        self.model.notes = values.get("Notes", self.model.notes)
        self.model.update()

    def doInsert(self, values):
        """
        Insert the Scope represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by MultipleScopeView or ScopeView containg all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        # Only multi insert exists at the moment for Scope
        # Get form values
        wave = values["wave"]
        ret = []
        total = 0
        accepted = 0
        insert_setting = values["Settings"]
        split_range_setting = values.get("Split", False)
        for line in values["Scopes"].split("\n"):
            if line.strip() != "":
                # Insert in database
                scopeToAdd = line.strip()
                if utils.isIp(scopeToAdd):
                    scopeToAdd += "/32"
                if utils.isNetworkIp(scopeToAdd):
                    if split_range_setting:
                        network_ips = utils.splitRange(scopeToAdd)
                        if len(network_ips) == 0:
                            model = Scope().initialize(wave, scopeToAdd, "")
                            inserted_res, iid = model.addInDb()
                        else:
                            for network_ip in network_ips:
                                model = Scope().initialize(wave,  str(network_ip), "")
                                inserted_res, iid = model.addInDb()
                                if inserted_res:
                                    accepted += 1
                                    ret.append(iid)
                                total += 1
                    else:
                        model = Scope().initialize(wave,  scopeToAdd, "")
                        inserted_res, iid = model.addInDb()
                else:
                    model = Scope().initialize(wave,  scopeToAdd, "")
                    inserted_res, iid = model.addDomainInDb(insert_setting)
                if inserted_res == 1:
                    accepted += 1
                    ret.append(iid)
                total += 1
        return ret, total-accepted  # nb errors = total - accepted

    
    def getChecks(self):
        """Return scope assigned checks as a list of checkInstance
        Returns:
            list of checkInstance objects
        """
        return self.model.getChecks()

    def getType(self):
        """Returns a string describing the type of object
        Returns:
            "scope" """
        return "scope"


"""Help the user to create a new pentest database.
"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.application.dialogs.ChildDialogNewPentest import ChildDialogNewPentest
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
import pollenisatorgui.core.components.utils as utils
from datetime import datetime
from PIL import Image, ImageTk
from pollenisatorgui.core.application.paginable import Paginable

class ChildDialogPentests(CTkToplevel):
    """
    Open a child dialog of a tkinter application to ask details about
    a new pentest database to create.
    """

    def __init__(self, parent, pentests):
        """
        Open a child dialog of a tkinter application to ask details about
        the new pentest.

        Args:
            parent: the tkinter parent view to use for this window construction.
        """
        super().__init__(parent)
        self.attributes("-type", "dialog")
        self.parent = parent
        self.pentests = pentests
        self.title("Pentest manager")
        self.resizable(True, True)
        monitor = utils.get_screen_where_widget(parent)
        self.geometry(f"{920}x{600}+{monitor.x+(monitor.width//2)-(920//2)}+{monitor.y+(monitor.height//2)-(600//2)}")
        
        self.rvalue = None
        
        self.bind("<Escape>", self.quit)
        self.link_image = CTkImage(Image.open(utils.getIcon("link.png"))) # button with text
        self.delete_image = tk.PhotoImage(file=utils.getIcon("delete.png")) # icon button
        # Load the image
        image=Image.open(utils.getIcon("copy_small.png"))
        self.copy_image = ImageTk.PhotoImage(image)
        image=Image.open(utils.getIcon("save_small.png"))
        self.export_image = ImageTk.PhotoImage(image)

        self.mainFrame = ScrollableFrameXPlateform(self)
        self.reloadUI()
        self.mainFrame.pack(fill=tk.BOTH, ipadx=10, ipady=10, expand=1)
        try:
            self.wait_visibility()
            self.transient(self.parent)
            self.focus_force()
            self.grab_set()
            self.lift()

        except tk.TclError:
            pass

    def reloadUI(self, reload_pentests=False):
        for widget in self.mainFrame.winfo_children():
            widget.destroy()
        if self.pentests is None or reload_pentests:
            apiclient = APIClient.getInstance()
            self.pentests = apiclient.getPentestList()
        else:
            self.pentests = self.pentests
        self.pentests.sort(key=lambda x: x["creation_date"], reverse=True)
        self.form = FormPanel()
        form1 = self.form.addFormPanel(fill=tk.BOTH)
        form1.addFormLabel("My Pentests")
        self.plus_image = CTkImage(Image.open(utils.getIcon("plus.png")))
        self.import_image = CTkImage(Image.open(utils.getIcon("import.png")))
        form1.addFormButton("Create new Pentest", self.selectNewPentest, image=self.plus_image, side=tk.LEFT)
        form1.addFormButton("Import archive", self.importPentest,  side=tk.LEFT, image=self.import_image)
        form1.addFormButton("Disconnect", self.disconnect,  side=tk.LEFT, padx=10)
        self.form.addFormSeparator()
        self.searchBar = self.form.addFormStr("Search", placeholder="Filter", binds={"<KeyRelease>": self.search}, side=tk.TOP, padx=10)
        self.form2 = self.form.addFormPanel(grid=True, fill=tk.BOTH, make_uniform_column=4)
        self.form2.addFormLabel("Pentest name", text="Pentest name", bold=True, row=0, column=0, sticky=tk.W)
        self.form2.addFormLabel("Creator", text="Creator", bold=True, row=0, column=1, sticky=tk.W)
        self.form2.addFormLabel("Date", text="Date", bold=True, row=0, column=2, sticky=tk.W)
        self.form2.addFormLabel("Actions", text="Actions", bold=True, row=0, column=3, sticky=tk.W)
        self.form2.addFormSeparator(row=1, column=0, columnspan=4, sticky=tk.EW)
        self.paginated_pentests = Paginable(self.mainFrame, self.callback_insert, self.callback_reset, self.callback_get_value, self.redraw)
        for pentest in self.pentests:
            pentest["iid"] = pentest["uuid"]
            self.paginated_pentests.addPaginatedInfo(pentest, False)
        content_view = self.paginated_pentests.getContentView()
        self.form.constructView(content_view)
        self.paginated_pentests.pack(fill=tk.BOTH, ipadx=10, ipady=10, expand=1)
        self.mainFrame.pack(fill=tk.BOTH, ipadx=10, ipady=10, expand=1)
        self.paginated_pentests.setPaginationPanel()
        try:
            self.wait_visibility()
            self.transient(self.parent)
            self.focus_force()
            self.grab_set()
            self.lift()
        except tk.TclError:
            pass

    def redraw(self):
        self.form2.redraw()

    def callback_insert(self, items):
        row = len(self.paginated_pentests.getShownInfos())*2 
        for pentest in items:
            date = pentest["creation_date"]
            date_o = datetime.fromisoformat(date)
            self.form2.addFormLabel("pentest_name", text=pentest["nom"],row=row, column=0, sticky=tk.W)
            self.form2.addFormLabel("pentest_creator", text=pentest["owner"], row=row, column=1, sticky=tk.W)
            self.form2.addFormLabel("pentest_date", text=utils.dateToString(date_o), row=row, column=2, sticky=tk.W)
            formAction = self.form2.addFormPanel(row=row, column=3, sticky=tk.W)
            formAction.addFormButton("", self.exportPentest, style="icon.TButton", side=tk.RIGHT, image=self.export_image, tooltip="Export", infos={"name":pentest["nom"], "uuid":pentest["uuid"]})
            formAction.addFormButton("", self.wrapCopyDb, style="icon.TButton", side=tk.RIGHT, image=self.copy_image, tooltip="duplicate pentest", infos={"name":pentest["nom"], "uuid":pentest["uuid"]})
            formAction.addFormButton("", self.deleteAPentest, style="icon.TButton", side=tk.RIGHT, image=self.delete_image, tooltip="delete pentest" , infos={"name":pentest["nom"], "uuid":pentest["uuid"]})
            formAction.addFormButton("Open", self.openPentest, width=55,image=self.link_image, side=tk.RIGHT, infos={"name":pentest["nom"], "uuid":pentest["uuid"]})
            self.form2.addFormSeparator(row=row+1, column=0, columnspan=4, sticky=tk.EW)
            row+=2
        

    def callback_reset(self):
        self.form2.clear()
        self.form2.addFormLabel("Pentest name", text="Pentest name", bold=True, row=0, column=0, sticky=tk.W)
        self.form2.addFormLabel("Creator", text="Creator", bold=True, row=0, column=1, sticky=tk.W)
        self.form2.addFormLabel("Date", text="Date", bold=True, row=0, column=2, sticky=tk.W)
        self.form2.addFormLabel("Actions", text="Actions", bold=True, row=0, column=3, sticky=tk.W)
        self.form2.addFormSeparator(row=1, column=0, columnspan=4, sticky=tk.EW)

    def callback_get_value(self, item, arg_n):
        columns = ["nom", "owner", "creation_date"]
        return item[columns[arg_n]]

    def search(self, event=None):
        """
        Search in treeviews
        """
        if event is not None:
            if event.keysym != "BackSpace" and event.keysym != "Delete" and (len(event.keysym) > 1 or not event.keysym.isalnum()):
                return
        search = self.searchBar.getValue()
        self.paginated_pentests.filter(search, search, search, check_all=False, check_case=False)


    def disconnect(self, event=None):
        apiclient = APIClient.getInstance()
        apiclient.disconnect()
        self.quit()

    def openPentest(self, event, infos):
        name = infos.get("uuid")
        self.rvalue = name
        self.quit()
    
    def deleteAPentest(self, event, infos):
        """
        Ask a user a pentest name then delete it.
        """
        apiclient = APIClient.getInstance()
        pentest = infos.get("uuid")
        pentestName = infos.get("name")
        dialog = ChildDialogQuestion(self, "Pentest deletion confirmation", "You are going to delete permanently the database \""+pentestName+"\". Are you sure ?")
        self.wait_window(dialog.app)
        if dialog.rvalue == "Yes":
            res = apiclient.doDeletePentest(pentest)
            if res:
                try:
                    path = os.path.join(utils.getLocalDir(), "states/", "."+pentestName)
                    os.remove(path)
                except FileNotFoundError:
                    pass
            self.reloadUI(reload_pentests=True)
        


    def selectNewPentest(self, _event=None):
        """
        Ask a user for a new pentest name. Then creates it.
        """
        validPentest = False
        default = {}
        while not validPentest:
            dialog = ChildDialogNewPentest(None, default)
            dialog.wait_window()
            if dialog.rvalue is None:
                return
            if isinstance(dialog.rvalue, dict):
                default = dialog.rvalue
                pentest_name = dialog.rvalue["name"]
                mission_name = dialog.rvalue["mission_name"]
                client_name = dialog.rvalue["client_name"]
                lang = dialog.rvalue["lang"]
                pentest_type = dialog.rvalue["type"]
                start_date = dialog.rvalue["start"]
                end_date = dialog.rvalue["end"]
                scope = dialog.rvalue["scope"]
                settings = dialog.rvalue["settings"]
                settings["mission_name"] = mission_name
                settings["client_name"] = client_name
                settings["lang"] = lang
                pentesters = dialog.rvalue["pentesters"]
                validPentest, msg = self.newPentest(pentest_name, pentest_type, start_date, end_date, scope, settings, pentesters)
                if validPentest:
                    self.rvalue = msg
                    self.quit()

    def newPentest(self, pentestName, pentest_type, start_date, end_date, scope, settings, pentesters):
        """
        Register the given pentest name into database and opens it.

        Args:
            pentestName: The pentest database name to register in database.
        """
        succeed = False
        if pentestName is not None:
            apiclient = APIClient.getInstance()
            succeed, msg = apiclient.registerPentest(pentestName, pentest_type, start_date, end_date, scope, settings, pentesters)
            if not succeed:
                tk.messagebox.showinfo("Forbidden", msg)
        return succeed, msg
            

    def quit(self, _event=None):
        self.destroy()

    def wrapCopyDb(self, _event, infos):
        """
        Call default copy database from a callback event.

        Args:
            _event: not used but mandatory
        """
        apiclient = APIClient.getInstance()
        pentestName = infos.get("uuid")
        success = apiclient.setCurrentPentest(pentestName)
        if not success:
            tk.messagebox.showerror("Failed", "Failed to connect to the selected pentest")
            return
        dialog = ChildDialogAskText(self, "Enter copy's name  of "+apiclient.getCurrentPentest()+" database name", multiline=False)
        self.wait_window(dialog.app)
        name = dialog.rvalue
        if name is None:
            return
        apiclient.copyDb(apiclient.getCurrentPentest(), name)
        self.reloadUI(reload_pentests=True)

    def importPentest(self, event=None):
        """
        Import a pentest archive file gunzip to database.
        Args:
            name: The filename of the gunzip database exported previously
        """
        apiclient = APIClient.getInstance()
        filename = ""
        f = tk.filedialog.askopenfilename(parent=self, defaultextension=".gz")
        if not isinstance(f, str):  # asksaveasfile return `None` if dialog closed with "cancel".
            return
        if len(str(f).strip()) == 0:
            return
        filename = str(f)
        original_dbname = ChildDialogAskText(self, "Enter the original database name", multiline=False)
        self.wait_window(original_dbname.app)
        if original_dbname.rvalue is None:
            return
        
        success = apiclient.importDb(filename, original_dbname.rvalue)
        if success:
            tk.messagebox.showinfo("Database import ", "Database import suceeded")
        else:
            tk.messagebox.showerror("Database import ", "Database import failed")
        self.reloadUI(reload_pentests=True)

    def exportPentest(self, event, info):
        """
        Dump a pentest database to an archive file gunzip.
        """
        apiclient = APIClient.getInstance()
        pentestName = info.get("uuid")
        success = apiclient.setCurrentPentest(pentestName)
        if not success:
            tk.messagebox.showerror("Failed", "Failed to connect to the selected pentest")
            return
        success, msg = apiclient.dumpDb(self, pentestName)
        if success is None:
            # closed
            return
        if not success:
            tk.messagebox.showerror("Database export error", msg)
        else:
            tk.messagebox.showinfo("Database export completed", msg)


"""View for ip object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.controllers.checkinstancecontroller import CheckInstanceController
from pollenisatorgui.core.controllers.scopecontroller import ScopeController
from pollenisatorgui.core.models.port import Port
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.models.ip import Ip
from pollenisatorgui.core.models.scope import Scope
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
from pollenisatorgui.core.views.multipleipview import MultipleIpView
from pollenisatorgui.core.views.multiplescopeview import MultipleScopeView
from pollenisatorgui.core.views.portview import PortView
from pollenisatorgui.core.views.defectview import DefectView
from pollenisatorgui.core.controllers.portcontroller import PortController
from pollenisatorgui.core.controllers.defectcontroller import DefectController
from pollenisatorgui.core.controllers.ipcontroller import IpController
from pollenisatorgui.core.views.viewelement import ViewElement
import pollenisatorgui.core.components.utils as utils
from tkinter import TclError
from bson.objectid import ObjectId
import json


class IpView(ViewElement):
    """View for ip object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory."""
    icon = 'ip.png'
    icon_out_of_scope = "ip_oos.png"
    cachedClassOOSIcon  = None

    def getIcon(self):
        if not self.controller.is_in_scope():
            from PIL import Image, ImageTk
            if self.__class__.cachedClassOOSIcon == None:
                path = utils.getIcon(self.__class__.icon_out_of_scope)
                self.__class__.cachedClassOOSIcon = ImageTk.PhotoImage(Image.open(path))
            return self.__class__.cachedClassOOSIcon
        return super().getIcon()

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Ip
        """
        self.form.clear()
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Ip")
        top_panel.addFormStr(
            "Ip", '', modelData["ip"], None, column=1, state="readonly")
        notes = modelData.get("notes", "")
        top_panel = self.form.addFormPanel()
        top_panel.addFormLabel("Notes", side="top")
        top_panel.addFormText("Notes", r"", notes, None, side="top")
        top_panel.addFormLabel("Infos", side="left")
        top_panel.addFormText("Infos", utils.is_json, json.dumps(modelData["infos"], indent=4, cls=utils.JSONEncoder), height=100, side="left", fill="both")
        buttons_panel = self.form.addFormPanel(grid=True)
        buttons_panel.addFormButton("Add a port", self.addPortCallback)
        buttons_panel.addFormButton(
            "Add a security defect", self.addDefectCallback, column=1)
        self.completeModifyWindow()

    def openInsertWindow(self):
        view = MultipleIpView(self.appliTw, self.appliViewFrame, self.mainApp, self.controller)
        view.openInsertWindow()

    def addPortCallback(self, _event=None):
        """
        Create an empty port model and its attached view. Open this view insert window.

        Args:
            _event: Automatically generated with a button Callback, not used.
        """
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()
        modelData = self.controller.getData()
        pv = PortView(self.appliTw, self.appliViewFrame,
                      self.mainApp, PortController(Port(modelData)))
        pv.openInsertWindow()

    def getAdditionalContextualCommands(self):
        if self.controller.is_in_scope():
            return {"Add IPs/Hosts": self.addAHostCallback,  "Add a port":self.addPortCallback, "Add a defect":self.addDefectCallback}
        else:
            return {"Add in scope": self.addToScopeCallback}
    
    def addAHostCallback(self, _event=None):
        objView = MultipleIpView(self.appliTw, self.appliViewFrame, self.mainApp, IpController(Ip()))
        objView.openInsertWindow()

    def addToScopeCallback(self, _event=None):
        objView = MultipleScopeView(self.appliTw, self.appliViewFrame, self.mainApp, ScopeController(Scope({"wave":"Main","scope":self.controller.model.ip})))
        objView.openInsertWindow(check_scope=False)

    def addDefectCallback(self, _event=None):
        """
        Create an empty defect model and its attached view. Open this view insert window.

        Args:
            _event: Automatically generated with a button Callback, not used.
        """
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()
        modelData = self.controller.getData()
        modelData["target_type"] = self.controller.model.__class__.__name__.lower()
        modelData["target_id"] = self.controller.model.getId()
        dv = DefectView(self.appliTw, self.appliViewFrame, self.mainApp,
                        DefectController(Defect(modelData)))
        dv.openInsertWindow(modelData.get("notes", ""))
        

    def _insertChildrenDefects(self):
        """Insert every children defect in database as DefectView under this node"""
        defects = self.controller.getDefects()
        for defect in defects:
            defect_o = DefectController(Defect(defect))
            defect_vw = DefectView(
                self.appliTw, self.appliViewFrame, self.mainApp, defect_o)
            defect_vw.addInTreeview(str(self.controller.getDbId()), addChildren=False)
        return defects
    
    def _insertChildrenChecks(self):
        """Create a tools list node and insert every children tools in database as ToolView under this node"""
        checks = self.controller.getChecks()
        for check in checks:
            check_o = CheckInstanceController(check)
            check_vw = CheckInstanceView(
                self.appliTw, self.appliViewFrame, self.mainApp, check_o)
            check_vw.addInTreeview(str(self.controller.getDbId()), addChildren=False)
        return checks
    
    def _insertChildrenPorts(self, ip_node):
        """Insert every children port in database as DefectView under this node directly"""
        ports = self.controller.getPorts()
        PortView.multiAddInTreeview(self.appliTw, self.appliViewFrame, self.mainApp, ports, ip_node)
    
    def _insertChildren(self):
        self._insertChildrenPorts(str(self.controller.getDbId()))
        self._insertChildrenDefects()
        self._insertChildrenChecks()

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            _addChildren: not used here
        """
        addChildren = kwargs.get("addChildren", True)
        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}

        if parentNode is None:
            parentNode = self.getParentNode()
        ip_node = None
        tags = self.controller.getTags()
        try:
            if isinstance(parentNode, ObjectId):
                ip_parent_o = Ip.fetchObject({"_id": parentNode})
                if ip_parent_o is not None:
                    parent_view = IpView(self.appliTw, self.appliViewFrame,
                                         self.mainApp, IpController(ip_parent_o))
                    parent_view.addInTreeview(None, addChildren=False)
            #CALLING TK CALL IS FASTER than
            #ip_node = self.appliTw.insert(parentNode, "end", str(
            #   self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=tags, image=self.getIcon())
            ip_node = self.appliTw.tk.call(self.appliTw._w, "insert", parentNode, "end", "-id", str(self.controller.getDbId()), 
                                 "-text", str(self.controller.getModelRepr()), "-tags", tags, "-image", self.getIcon())
        except TclError as e:
            pass
        if addChildren and ip_node is not None:
            self._insertChildren()
        elif not addChildren and self.appliTw.lazyload and not self.mainApp.searchMode:
            try:
                self.appliTw.tk.call(self.appliTw._w, "insert", ip_node, "end", "-id", str(self.controller.getDbId())+"|<Empty>", 
                                 "-text", "<Empty>")
            except TclError as e:
                pass
        # self.appliTw.sort(parentNode)
        if "hidden" in tags:
            self.hide("tags")
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")
        modelData = self.controller.getData()
        if self.mainApp.settings.is_hide_oos() and not modelData["in_scopes"]:
            self.hide("filter_oos")
        if not modelData["in_scopes"]:
            # calling tk.call faster than 
            #self.appliTw.item(ip_node, tags=tags+["OOS"], image=self.getIcon())
            self.appliTw.tk.call(
                self.appliTw._w, "item", ip_node, "-tags", tags+["OOS"], "-image", self.getIcon())
            

    def split_ip(self):
        """Split a IP address given as string into a 4-tuple of integers.
        Returns:
            Tuple of 4 integers values representing the 4 parts of an ipv4 string"""
        modelData = self.controller.getData()
        try:
            ret = tuple(int(part) for part in modelData["ip"].split('.'))
        except ValueError:
            ret = tuple([0])+tuple(ord(chrDomain)
                                   for chrDomain in modelData["ip"])
        return ret

    def key(self):
        """Returns a key for sorting this node
        Returns:
            Tuple of 4 integers values representing the 4 parts of an ipv4 string, key to sort ips properly
        """
        return self.split_ip()

    def insertReceived(self):
        """Called when a IP insertion is received by notification.
        Insert the node in summary.
        Can also insert in treeview with OOS tags.
        """
        modelData = self.controller.getData()
        if modelData.get("in_scopes", []): # in_scopes is not empty
            for module in self.mainApp.modules:
                if callable(getattr(module["object"], "insertIP", None)):
                    module["object"].insertIp(modelData["ip"])
        else:
            self.appliTw.item(str(self.controller.getDbId()), text=str(
                self.controller.getModelRepr()), image=self.getIcon(), tags=["OOS"])

    def updateReceived(self, obj=None, old_obj=None):
        """Called when a IP update is received by notification.
        Update the ip node OOS status tags and add/remove it from summary.
        """
        if self.controller.model is not None:
            modelData = self.controller.getData()
            if not modelData["in_scopes"]:
                self.controller.addTag("OOS")
                self.appliTw.item(str(self.controller.getDbId()), image=self.getIcon(), tags=self.controller.getTags()+["OOS"])
                for module in self.mainApp.modules:
                    if callable(getattr(module["object"], "deleteIp", None)):
                        module["object"].deleteIp(modelData["ip"])
            else:
                tags = list(self.controller.getTags())
                if "OOS" in tags:
                    tags.remove("OOS")
                self.appliTw.item(str(self.controller.getDbId()), tags=tags, image=self.getIcon())
                for module in self.mainApp.modules:
                    if callable(getattr(module["object"], "insertIp", None)):
                        module["object"].insertIp(modelData["ip"])
            super().updateReceived()

    def getParentNode(self):
        """
        Return the id of the parent node in treeview.

        Returns:
            return the parent ips_node of application treeview
        """
        #parent = self.controller.getParent()
        #if parent is None:
        parent = self.appliTw.ips_node
        return parent



"""Widget "Table" using a ttk Treeview"""

import tkinter as tk
import tkinter.ttk as ttk
import tkinter.font as tkfont
from customtkinter import *
from pollenisatorgui.core.forms.form import Form
import pyperclip
from pollenisatorgui.core.application.dialogs.ChildDialogCombo import ChildDialogCombo
from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
import pollenisatorgui.core.components.utils as utils

class FormTreevw(Form):
    """
    Form field representing a multi-lined input.
    """

    def __init__(self, name, headings, default_values=None, **kwargs):
        """
        Constructor for a form text

        Args:
            name: the treeview name (id).
            headings: table headers
            default_values: default values for the Table as a dict, default is None
            kwargs: same keyword args as you would give to ttk.Treeview
                    Plus:
                        - doubleClickBinds : list of values to edit columns as:
                                        - None : Not editable
                                        - string : ask user to input a string
                                        - list : ask user to choose in a combo box one value in the list 

        """
        super().__init__(name)
        self.headings = headings
        self.default_values = default_values if default_values is not None else {}
        self.treevw = None
        self.kwargs = kwargs
        self.tvFrame = None
        self.scbVSel = None
        self.widgetMenuOpen = None
        self.contextualMenu = kwargs.get("contextualMenu", None)
        self.doubleClickBinds = kwargs.get("doubleClickBinds", None)
        self.f = None
        self._detached = set()
        self._cache = {}
        self.type = "dict"
        self.movingSelection = None
        self.lastMovedTo = None

    def _initContextualMenu(self, parent):
        """Initialize the contextual menu for paperclip.
        Args:
            parent: the tkinter parent widget for the contextual menu
        """
        self.contextualMenu = utils.craftMenuWithStyle(parent)
        parent.bind("<Button-3>", self.popup)
        self.contextualMenu.add_command(label="Copy", command=self.copy)
        self.contextualMenu.add_command(label="Close", command=self.close)

    def addContextMenuCommand(self, label, command, replace=False):
        found = False
        if self.contextualMenu is None:
            self._initContextualMenu(self.treevw)
        for i in range(self.contextualMenu.index('end')+1):
            labelStr = str(self.contextualMenu.entrycget(i,'label') )
            if labelStr == label and not replace:
                found = True
                break
        if not found:
            self.contextualMenu.add_command(label=label, command=command)

    def close(self):
        """Option of the contextual menu : Close the contextual menu by doing nothing
        """
        pass

    def copy(self, _event=None):
        """Option of the contextual menu : Copy entry text to clipboard
        """
        selected = self.treevw.selection()
        texts = []
        for item in selected:
            it = self.treevw.item(item)
            texts.append(it.get("text", "") + " " +
                         " ".join(it.get("values", [])))

        pyperclip.copy("\n".join(texts))

    def popup(self, event):
        """
        Fill the self.widgetMenuOpen and reraise the event in the editing window contextual menu

        Args:
            event: a ttk Treeview event autofilled.
            Contains information on what treeview node was clicked.
        """
        self.widgetMenuOpen = event.widget
        self.contextualMenu.tk_popup(event.x_root, event.y_root)
        self.contextualMenu.focus_set()
        self.contextualMenu.bind('<FocusOut>', self.popupFocusOut)

    def popupFocusOut(self, _event=None):
        """Callback for focus out event. Destroy contextual menu
        Args:
            _event: not used but mandatory
        """
        self.contextualMenu.unpost()

    def recurse_insert(self, values, parent='', columnsLen=None, odd=False):
        """Recursive insert of a value in a table:
        Args:
            values: values to insert in the treeview
                    * If it is a dict : Recurse
                    * If it is a list : Add key without value and list values as subchildren
                    * If it is a str : Insert into parent
                    * If it is a tuple : Insert first value as text and other values as values in the same line
            parent: the parent node treeview id to insert values into
            columnsLen: a table with the width of each column as list of 2 int
            odd: insert value as an odd value (the line will be tagged odd and change color). Default is False
        Returns:
            Final size of columns as list of two int
        """
        child_odd = False
        if columnsLen is None:
            columnsLen = [self.f.measure(str(x)) for x in self.headings]
        if isinstance(values, dict):
            # treeview
            self.type = "dict"
            sorted_keys = sorted(list(values.keys()))
            for sorted_key in sorted_keys:
                key = sorted_key
                value = values[key]
                root = parent
                if isinstance(value, str):
                    self.treevw.insert(root, tk.END, None, text=key, values=[
                        str(value)], tags=("odd") if odd else ())
                    odd = not odd
                    columnsLen[0] = max(columnsLen[0], self.f.measure(key))
                    columnsLen[1] = min(
                        1000, max(columnsLen[1], self.f.measure(str(value))))
                elif isinstance(value, list):
                    if parent != '':
                        return 
                    root = self.treevw.insert(
                        root, tk.END, None, text=key, tags=("odd") if odd else ())
                    odd = not odd
                    columnsLen[0] = max(columnsLen[0], self.f.measure(key))
                    child_odd = False
                    for listValue in value:
                        self.treevw.insert(root, tk.END, None,
                                        text="", values=[str(listValue)], tags=("odd") if child_odd else ())
                        columnsLen[1] = min(1000, max(
                            columnsLen[1], self.f.measure(str(listValue))))
                        child_odd = not child_odd
                elif isinstance(value, tuple):
                    if len(value) == 0:
                        continue
                    value_arr = list(value[1:]) if len(value) > 1 else []
                    self.treevw.insert(root, tk.END, None, text=str(value[0]), values=value_arr, tags=("odd") if child_odd else ())
                    for i in range(len(value)):
                        try:
                            columnsLen[i] = min(1000, max(
                                columnsLen[i], self.f.measure(str(value[i]))))
                        except IndexError:
                            columnsLen.insert(i, min(1000, self.f.measure(str(value[i]))))
                    child_odd = not child_odd
                elif isinstance(value, dict):
                    parent = self.treevw.insert(
                        root, tk.END, None, text=key, tags=("odd") if odd else ())
                    odd = not odd
                    columnsLen[0] = max(columnsLen[0], self.f.measure(key))
                    columnsLen[1] = min(1000, max(
                        columnsLen[1], self.f.measure(str(value))))
                    columnsLen = self.recurse_insert(
                        value, parent=root, columnsLen=columnsLen, odd=odd)
        elif isinstance(values, list):
            # table like
            self.type = "list"
            for sub_list in values:
                if len(sub_list) == 0:
                        continue
                value_arr = list(sub_list[1:]) if len(sub_list) > 1 else []
                self.treevw.insert(parent, tk.END, None, text=str(sub_list[0]), values=value_arr, tags=("odd") if child_odd else ())
                for i in range(len(sub_list)):
                    try:
                        columnsLen[i] = min(1000, max(
                            columnsLen[i], self.f.measure(str(sub_list[i]))))
                    except IndexError:
                        columnsLen.insert(i, min(1000, self.f.measure(str(sub_list[i]))))
                child_odd = not child_odd
        return columnsLen

    def constructView(self, parent):
        """
        Create the text view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        from pollenisatorgui.core.components.settings import Settings
        w = self.getKw("width", 600)
        self.tvFrame = CTkFrame(
            parent.panel, width=w, height=0)
        h = self.getKw("height", len(self.default_values)+1)
        mh = self.getKw("max_height", 10)
        self.treevw = ttk.Treeview(self.tvFrame, height=min(h, mh))
        settings = Settings()
        font = tkfont.nametofont('TkTextFont')
        self.treevw.tag_configure("even", background=utils.getBackgroundColor(), font=tkfont.nametofont('TkTextFont'))
        self.treevw.tag_configure("odd", background=utils.getBackgroundSecondColor(), font=tkfont.nametofont('TkTextFont'))
        self.scbVSel = CTkScrollbar(self.tvFrame,
                                     orientation=tk.VERTICAL,
                                     command=self.treevw.yview, height=0)
        scbHSel = CTkScrollbar(
            self.tvFrame, orientation=tk.HORIZONTAL, command=self.treevw.xview)
        self.treevw.configure(yscrollcommand=self.scbVSel.set)
        self.treevw.configure(xscrollcommand=scbHSel.set)
        self.tvFrame.rowconfigure(0, weight=1)
        self.tvFrame.columnconfigure(0, weight=1)

        self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
        self.scbVSel.grid(row=0, column=1, sticky=tk.NS)
        scbHSel.grid(row=1, column=0, sticky=tk.EW)
        if len(self.headings) > 1:
            self.treevw['columns'] = self.headings[1:]
        root = CTkLabel(self.tvFrame)
        self.f = tkfont.nametofont('TkTextFont')
        columnsLen = self.recurse_insert(self.default_values)
        listOfLambdas = [self.column_clicked("#"+str(i), False) for i in range(len(self.headings))]
        autoresize = self.getKw("auto_size_columns", True)
        for h_i, header in enumerate(self.headings):
            self.treevw.heading("#"+str(h_i), text=header, anchor="w", command=listOfLambdas[h_i])
            if autoresize:
                self.treevw.column("#"+str(h_i), anchor='w',
                               stretch=tk.YES, minwidth=columnsLen[h_i], width=columnsLen[h_i])
        
        emptyRowFound = False
        add_empty_row = self.getKw("add_empty_row", False)
        children = self.treevw.get_children()
        for child in children:
            item = self.treevw.item(child)
            if item["text"] == "":
                emptyRowFound = True
                break
        if self.getKw("status", "none") != "readonly":
            if not emptyRowFound and add_empty_row:
                tags = ("odd") if len(children) % 2 == 1 else ()
                self.treevw.insert('', tk.END, None, text="",
                                values=["" for x in range(len(self.headings)-1)], tags=tags)
            self.treevw.bind("<Double-Button-1>", self.OnDoubleClick)
            self.treevw.bind("<Delete>", self.deleteItem)
            self.treevw.bind("<Control-c>", self.copy)
            self.treevw.bind("<Alt-Down>",self.bDown)
            self.treevw.bind("<Alt-Up>",self.bUp)
            self.treevw.bind("<ButtonPress-1>",self.dragStart)
            self.treevw.bind("<ButtonRelease-1>",self.dragRelease, add='+')
            self.treevw.bind("<B1-Motion>",self.dragMove, add='+')
        self._initContextualMenu(self.treevw)
        binds = self.getKw("binds", {})
        for key, val in binds.items():
            self.treevw.bind(key, val)
        if parent.gridLayout:
            self.tvFrame.grid(row=self.getKw("row", 0), column=self.getKw(
                "column", 0), sticky=self.getKw("sticky", tk.NSEW), fill=self.getKw("fill", "both"), expand=self.getKw("expand", False))

        else:
            self.tvFrame.pack(side=self.getKw("side", "top"), anchor=self.getKw("anchor", None), padx=self.getKw(
                "padx", 10), pady=self.getKw("pady", 5), fill=self.getKw("fill", "both"), expand=self.getKw("expand", False))
        
    def column_clicked(self, col, reverse):
        """A lambda to call the statusbarController.statusbarClicked with the tag name clicked
        Args:
            name: the tag name clicked
        """
        return lambda : self.sort_column(self.treevw, col, reverse)
    
    def filter(self, *args, **kwargs):
        if kwargs.get("reset", True):
            children = list(self._detached) + list(self.treevw.get_children())
            self._detached = set()
        else:
            children = list(self.treevw.get_children())
        self._brut_searcher(children, *args, **kwargs)

    def _brut_searcher(self, children, *args, **kwargs):
        i_r = -1
        for item_id in children:
            allValid = True
            oneValid = False
            for iarg, arg in enumerate(args):
                if iarg == 0:
                    text = self.treevw.item(item_id)['text']
                else:
                    text = self.treevw.item(item_id)['values'][iarg-1]
                check_all = kwargs.get("check_all", True)
                if isinstance(arg, str):
                    is_valid = arg in text
                elif isinstance(arg, bool):
                    is_valid = arg
                elif isinstance(arg, list):
                    is_valid = text in arg
                if not is_valid:
                    allValid = False
                    if check_all:
                        break
                else:
                    oneValid = True
                    if not check_all:
                        break
            if allValid and check_all:
                i_r += 1
                self.treevw.reattach(item_id, '', i_r)
            elif oneValid and not check_all:
                i_r += 1
                self.treevw.reattach(item_id, '', i_r)
            else:
                self._detached.add(item_id)
                self.treevw.detach(item_id)
        self.resetOddTags()

    def sort_column(self, tv, col, reverse):
        if col != "#0":
            l = [(tv.set(k, col), k) for k in tv.get_children('')]
        else:
            l = [(tv.item(k)["text"],k) for k in tv.get_children('')]
        l.sort(reverse=reverse)
        # rearrange items in sorted positions
        for index, (val, k) in enumerate(l):
            tv.move(k, '', index)
            if index % 2 != 0:
                tv.item(k, tags=("odd"))
            else:
                tv.item(k, tags=())
           

        # reverse sort next time
        tv.heading(col, command=self.column_clicked(col, not reverse))

    def reset(self):
        """Reset the treeview values (delete all lines)"""
        self._detached = set()
        if self.treevw is None:
            return
        for item in self.treevw.get_children():
            self.treevw.delete(item)

    def resetOddTags(self):
        for i, child in enumerate(self.treevw.get_children()):
            tags = ("odd") if i%2 != 0 else ()
            self.treevw.item(child, tags=tags)

    def deleteItem(self, _event=None):
        """Callback for <Del> event
        Remove the selected item in the treeview
        Args:
            _event: not used but mandatory"""
        for selected in self.treevw.selection():
            item = self.treevw.item(selected)
            if item["text"].strip() != "":
                self.treevw.delete(selected)
        self.resetOddTags()

    def addItem(self, parent="", insertPos="0", iid=None, **kwargs):
        self.treevw.insert(parent, insertPos, iid, **kwargs)
        self.resetOddTags()

    def setValue(self, values):
        self.reset()
        self.recurse_insert(values)

    def auto_resize_columns(self):
        columnsLen = [self.f.measure(str(x)) for x in self.headings]    
        for children in self.treevw.get_children():
            item = self.treevw.item(children)
            columnsLen[0] = max(columnsLen[0], self.f.measure(item["text"]))
            for i, value in enumerate(item["values"]):
                columnsLen[i+1] = max(columnsLen[i+1], self.f.measure(value))
        for h_i, header in enumerate(self.headings):
            self.treevw.column("#"+str(h_i), minwidth=columnsLen[h_i], width=columnsLen[h_i])

    def OnDoubleClick(self, event):
        """Callback for double click event
        Edit value of double clicked item
        Args:
            event: automatically filled when event is triggered.
        """
        item = self.treevw.identify("item", event.x, event.y)
        column = self.treevw.identify_column(event.x)
        columnNb = int(column[1:])
        values = self.treevw.item(item)["values"]
        if columnNb == 0:
            oldVal = self.treevw.item(item)["text"]
        else:
            try:
                oldVal = values[columnNb-1]
            except IndexError:
                oldVal = ""
        if self.doubleClickBinds is None:
            dialog = ChildDialogAskText(self.tvFrame, "New value for "+self.headings[columnNb].lower(), default=oldVal, multiline=True, width=100)
            self.tvFrame.wait_window(dialog.app)
            if dialog.rvalue is None:
                return
            newVal = dialog.rvalue
        else:
            binding = self.doubleClickBinds[columnNb]
            if binding is None:
                return
            elif callable(binding):
                newVal = binding(oldVal)
                if newVal is None:
                    return
            elif isinstance(binding, str):
                dialog = ChildDialogAskText(self.tvFrame, "New value for "+self.headings[columnNb].lower(), default=oldVal, multiline=True, width=100)
                self.tvFrame.wait_window(dialog.app)
                if dialog.rvalue is None:
                    return
                newVal = dialog.rvalue
            elif isinstance(binding, list):
                dialog = ChildDialogCombo(None, binding, "Choose one:")
                dialog.app.wait_window(dialog.app)
                newVal = dialog.rvalue
        if newVal is None:
            return
        if newVal.strip() == "" or newVal.strip() == oldVal.strip():
            return
        if columnNb == 0:
            self.treevw.item(item, text=newVal.strip())
        else:
            newVals = list(values)
            newVals[columnNb-1] = newVal.strip()
            self.treevw.item(item, values=newVals)
        emptyRowFound = False
        children = self.treevw.get_children()
        for child in children:
            item = self.treevw.item(child)
            if item["text"] == "":
                emptyRowFound = True
                break
        if not emptyRowFound:
            tags = ("odd") if len(children) % 2 == 1 else ()
            self.treevw.insert('', tk.END, None, text="",
                               values=["" for x in range(len(self.headings)-1)], tags=tags)
            currentHeight = len(self.treevw.get_children())
            if currentHeight < self.getKw("max_height", 5):
                self.treevw.configure(height=currentHeight)

    def selection(self):
        return self.treevw.selection()
    
    def item(self, iid, **kwargs):
        return self.treevw.item(iid, **kwargs)


    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the entry value as string.
        """
        if self.type == "dict":
            ret = {}
            children = self.treevw.get_children()
            for child in children:
                item = self.treevw.item(child)
                if not item["values"]:
                    children_list = self.treevw.get_children(child)
                    if children_list:
                        for child_list in children_list:
                            ret[item["text"]] = ret.get(item["text"], [])
                            item_child_list = self.treevw.item(child_list)
                            if item_child_list["values"]:
                                ret[item["text"]].append(item_child_list["values"][0])
                    else:
                        ret[item["text"]] = ""
                else:
                    ret[item["text"]] = item["values"]
            return ret
        elif self.type == "list":
            ret = []
            children = self.treevw.get_children()
            for child in children:
                item = self.treevw.item(child)
                elem = []
                elem.append(item["text"])
                elem += list(item["values"])
                ret.append(elem)
            return ret

    def bDown(self, event=None):
        item_iid = self.treevw.selection()[0]
        children = self.treevw.get_children()
        iid_moving = children.index(item_iid)
        try:
            iid_moved_by = children[iid_moving+1]
            self.treevw.move(item_iid, '', iid_moving+1)
        except IndexError:
            pass
        return "break"


    def bUp(self, event=None):
        item_iid = self.treevw.selection()[0]
        children = self.treevw.get_children()
        iid_moving = children.index(item_iid)
        try:
            iid_moved_by = children[iid_moving-1]
            self.treevw.move(item_iid, '', iid_moving-1)
        except IndexError:
            pass
        return "break"

    def dragStart(self, event):
        tv = event.widget
        if tv.identify_row(event.y) not in tv.selection():
            tv.selection_set(tv.identify_row(event.y))    
            self.movingSelection = tv.identify_row(event.y)

    def dragRelease(self, event):
        if self.movingSelection is None or self.lastMovedTo is None:
            return
        tv = event.widget
        if tv.identify_row(event.y) in tv.selection():
            self.movingSelection = None
            self.lastMovedTo = None

    def dragMove(self, event):
        tv = event.widget
        rowToMove = tv.identify_row(event.y)
        moveto = tv.index(rowToMove)    
        self.lastMovedTo = rowToMove if rowToMove != self.movingSelection else self.lastMovedTo
        for s in tv.selection():
            tv.move(s, '', moveto)


"""A collapsible panel form field."""

import tkinter.ttk as ttk
from customtkinter import *
import tkinter as tk
from pollenisatorgui.core.application.CollapsibleFrame import CollapsibleFrame

from pollenisatorgui.core.forms.formpanel import FormPanel


class FormCollapsbilePanel(FormPanel):
    """
    Form field representing a collapsible panel. It is composed of other forms.
    Additional kwargs values:
        grid: set the layout to grid, default to False
    Default setted values:
        width=500
        if pack : padx = 10, pady = 5, side = top, fill = x
        if grid: row = column = 0 sticky = "East"
    """

    def __init__(self, text, **kwargs):
        """
        Constructor for a panel.
        """
        super().__init__(**kwargs)
        self.text = text
        self.kwargs = kwargs
        self.gridLayout = self.getKw("grid", False)

    def constructView(self, parent):
        """
        Create the panel view by constructing all subforms views inside a tkinter panel the parent view given.
        Args:
            parent: parent view or parent FormPanel.
        """
        self.panel_frame = CollapsibleFrame(parent.panel, self.text, fg_color=self.getKw("fg_color", None), height=self.getKw("height", 20), 
                                      interior_padx=self.getKw("interior_padx", 0), interior_pady=self.getKw("interior_pady", 0))
        self.panel = self.panel_frame.interior
        self.populateView(parent)

    def populateView(self, parent):
        try:
            if self.make_uniform_column is not None:
                self.makeUniformColumn(self.make_uniform_column)
            
            for form in self.subforms:
                form.constructView(self)
            self.panel_frame.update_width()
            if isinstance(parent, FormPanel):  # Panel is a subpanel
                if parent.gridLayout:

                    self.panel_frame.grid(column=self.getKw(
                        "column", 0), row=self.getKw("row", 0), sticky=self.getKw("sticky", tk.NSEW), **self.kwargs)
                else:
                    self.panel_frame.pack(fill=self.getKw("fill", "both"), side=self.getKw(
                        "side", "top"), pady=self.getKw("pady", 5), padx=self.getKw("padx", 10),  **self.kwargs)
            else:  # Master panel, packing
                is_grid = "row" in self.kwargs or "column" in self.kwargs
                if not is_grid:
                    self.panel_frame.pack(fill=self.getKw("fill", "both"), side="top", pady=self.getKw("pady", 5), padx=self.getKw("padx", 30))
                else:
                    self.panel_frame.grid(sticky=self.getKw("sticky", tk.NSEW), row=self.getKw("row", 0), column=self.getKw("column", 0), pady=self.getKw("pady", 5), padx=self.getKw("padx", 30))
        except Exception as e:
            raise Exception("Error while populating view of panel named '" + self.name + "' : " + str(e))
            



"""View for multi checkinstance list object. Present a form to user when interacted with."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.viewelement import ViewElement
from customtkinter import *
from PIL import ImageTk, Image
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.components.settings import Settings
import tkinter.ttk as ttk
import tkinter as tk


class MultiTodoCheckInstanceView(ViewElement):
    """View for checlist multi object. Present an multi  form to user when interacted with."""

    ### TODO : CHECKLIST VIEW ADD SCOPES filtering when there is too many checkintances.
    def __init__(self, appliTw, appViewFrame, mainApp, models, parent=None):
        super().__init__(appliTw, appViewFrame, mainApp, None)
        self.checks = models
        self.parent = parent

    def addInTreeview(self, title, **kwargs):
        from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
        iid = "multi|todo|"+str(self.parent)
        self.appliTw.views[iid] = {"view": self}
        self.appliTw.insert(self.parent, "end", iid,text=f"{title} TODO ({len(self.checks)})", image=CheckInstanceView.getStatusIcon("todo"))

    def queueAll(self, _event=None):
        apiclient = APIClient.getInstance()
        apiclient.queueCheckInstances([str(c) for c in self.checks])

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or perform actions on multiple different objects common properties like tags.
        """
        self.form.clear()
        top_panel = self.form.addFormPanel(fill=tk.NONE)
        ready, msg = self.mainApp.scanManager.is_ready_to_queue()
        self.buttonQueueImage = CTkImage(Image.open(utils.getIcon('exec_cloud.png')))
        top_panel.addFormButton("Queue all", self.queueAll, state="normal" if ready else "disabled", image=self.buttonQueueImage)
        if not ready:
            top_panel.addFormLabel(msg)
        top_panel = self.form.addFormPanel(fill=tk.NONE)
        top_panel.addFormButton("Export", self.appliTw.exportSelection)
        self.delete_image = CTkImage(Image.open(utils.getIcon("delete.png")))
        #top_panel.addFormButton("Custom Command", self.appliTw.customCommand)
        top_panel.addFormButton("Delete", self.appliTw.deleteSelected, image=self.delete_image,
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        # panTags = self.form.addFormPanel(grid=True)
        # registeredTags = Settings.getTags()
        # keys = list(registeredTags.keys())
        # column = 0
        # listOfLambdas = [self.tagClicked(keys[i]) for i in range(len(keys))]
        # for registeredTag, tag_info in registeredTags.items():
        #     s = ttk.Style(self.mainApp)
        #     btn_tag = panTags.addFormButton(registeredTag, listOfLambdas[column], column=column)
        #     btn_tag.configure(fg_color=tag_info.get("color"))
        #     column += 1
        self.showForm()



"""
Module to add defects and export them
"""
import tkinter.ttk as ttk
from customtkinter import *
import tkinter as tk
import tkinter.messagebox
import os
import threading
from PIL import ImageTk, Image
from bson.objectid import ObjectId
from pollenisatorgui.core.components.logger_config import logger
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.models.defect import Defect
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.dialogs.ChildDialogCombo import ChildDialogCombo
from pollenisatorgui.core.application.dialogs.ChildDialogInfo import ChildDialogInfo
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.application.dialogs.ChildDialogToast import ChildDialogToast
from pollenisatorgui.core.application.dialogs.ChildDialogDefectView import ChildDialogDefectView
from pollenisatorgui.core.application.dialogs.ChildDialogRemarkView import ChildDialogRemarkView
from pollenisatorgui.core.models.remark import Remark	
from pollenisatorgui.core.views.remarkview import RemarkView
from pollenisatorgui.modules.module import Module

class Report(Module):
    """
    Store elements to report and create docx or xlsx with them
    """
    iconName = "tab_report.png"
    tabName = "    Report    "
    collNames = ["defects", "remarks"]
    order_priority = Module.HIGH_PRIORITY
    
    def __init__(self, _parent, settings, tkApp):

        """
        Constructor
        """
        super().__init__()
        self.tkApp = tkApp
        self.langs = ["en"]
        self.docx_models = []
        self.curr_lang = "en"
        self.pptx_models = []
        self.mainRedac = "N/A"
        self.settings = settings
        self.dragging = None
        self.parent = None
        self.reportFrame = None
        self.rowHeight = 0
        self.pane_base_height = 31
        self.style = None
        self.defect_treevw = None
        self.drag_toast = None
        self.combo_word = None
        self.combo_pptx = None
        self.btn_template_photo = None
        self.lastMovedTo = None
        self.movingSelection = None
        self.inited = False
        return

    def open(self, view, nbk, treevw):
        self.nbk = nbk
        self.main_app_treevw = treevw
        if self.inited is False:
            self.initUI(view)
        self.refreshUI()
        return True


    def refreshUI(self):
        """
        Reload informations and reload them into the widgets
        """
        self.langs = APIClient.getInstance().getLangList()
        self.combo_lang.configure(values=self.langs)
        self.combo_lang.set(self.settings.db_settings.get("lang", "en"))
        self.langChange(None)
        self.reset()
        self.fillWithDefects()
        self.fillWithRemarks()
        return
        

    def initUI(self, parent):
        """
        Initialize window and widgets.
        """
        if self.inited:  # Already initialized
            self.reset()
            self.fillWithDefects()
            self.fillWithRemarks()
            return
        self.inited = True
        self.parent = parent
        ### MAIN PAGE FRAME ###
        self.reportFrame = CTkScrollableFrame(parent)
        self.image_add = CTkImage(Image.open(utils.getIcon("plus.png")))
        self.image_del = CTkImage(Image.open(utils.getIcon("delete.png")))
        self.image_edit = CTkImage(Image.open(utils.getIcon("stylo.png")))
        self.image_docs = CTkImage(Image.open(utils.getIcon("documents.png")))
        self.image_word = CTkImage(Image.open(utils.getIcon("word.png")))
        self.image_ppt = CTkImage(Image.open(utils.getIcon("ppt.png")))
        ### DEFECT TABLE ###
        self.rowHeight = 20
        self.style = ttk.Style()
        self.style.configure('Report.Treeview', rowheight=self.rowHeight)
        # REMARK TREEVW	
        self.remarksFrame = ttk.LabelFrame(self.reportFrame, text="Remarks table")	
        self.paned_remarks = tk.PanedWindow(self.remarksFrame, orient=tk.VERTICAL, height=900)	
        self.remarkframeTw = CTkFrame(self.paned_remarks)	
        frameButtonsRemark = CTkFrame(self.remarkframeTw)
        self.remarks_treevw = ttk.Treeview(self.remarkframeTw, style='Report.Treeview', height=0)	
        self.remarks_treevw["columns"] = ["Title", "Type"]	
        self.remarks_treevw.heading("#0", text='Title', anchor=tk.W)	
        self.remarks_treevw.column("#0", width=150, anchor=tk.W)	
        self.remarks_treevw.bind("<Delete>", self.deleteSelectedRemarkItem)	
        self.remarks_treevw.bind("<Double-Button-1>", self.OnRemarkDoubleClick)
        self.remarks_treevw.grid(row=0, column=0, sticky=tk.NSEW)	
        scbVSel = CTkScrollbar(self.remarkframeTw,	
                                orientation=tk.VERTICAL,	
                                command=self.remarks_treevw.yview)	
        self.remarks_treevw.configure(yscrollcommand=scbVSel.set)	
        scbVSel.grid(row=0, column=1, sticky=tk.NS)	
        self.remarkframeTw.columnconfigure(0, weight=1)	
        self.remarkframeTw.rowconfigure(0, weight=1)	
        self.remarkframeTw.pack(side=tk.TOP, fill=tk.BOTH, pady=5, expand=1, padx=5)	
        btn_addRemark = CTkButton(	
            frameButtonsRemark, text="Add remark", image=self.image_add, command=self.addRemarkCallback)	
        btn_addRemark.pack(side=tk.TOP, anchor=tk.CENTER, pady=5)	
        btn_delRemark = CTkButton(	
            frameButtonsRemark, text="Remove selection", image=self.image_del, command=self.deleteSelectedRemarkItem,
                                    fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")	
        btn_delRemark.pack(side=tk.TOP, anchor=tk.CENTER, pady=5)	
        frameButtonsRemark.grid(row=0, column=2, sticky=tk.W + tk.E)	
        frameAllBelow = CTkFrame(self.paned_remarks)	
            
        # DEFECT TREEVW	
        defectLabelFrame = ttk.LabelFrame(frameAllBelow, text="Defects table")	
        self.paned = tk.PanedWindow(defectLabelFrame, orient=tk.VERTICAL, height=800)

        self.frameTw = CTkFrame(self.paned)
        self.treevw = ttk.Treeview(self.frameTw, style='Report.Treeview', height=0)
        self.treevw['columns'] = ('ease', 'impact', 'risk', 'type', 'redactor')
        self.treevw.heading("#0", text='Title', anchor=tk.W)
        self.treevw.column("#0", anchor=tk.W, width=150)
        self.treevw.heading('ease', text='Ease')
        self.treevw.column('ease', anchor='center', width=40)
        self.treevw.heading('impact', text='Impact')
        self.treevw.column('impact', anchor='center', width=40)
        self.treevw.heading('risk', text='Risk')
        self.treevw.column('risk', anchor='center', width=40)
        self.treevw.heading('type', text='Type')
        self.treevw.column('type', anchor='center', width=10)
        self.treevw.heading('redactor', text='Redactor')
        self.treevw.column('redactor', anchor='center', width=20)
        self.treevw.tag_configure(
            "Critical", background="black", foreground="gray97")
        self.treevw.tag_configure(
            "Major", background="red", foreground="gray97")
        self.treevw.tag_configure(
            "Important", background="orange", foreground="gray97")
        self.treevw.tag_configure(
            "Minor", background="yellow", foreground="black")
        self.treevw.bind("<Double-Button-1>", self.OnDoubleClick)
        self.treevw.bind("<Delete>", self.deleteSelectedItem)
        self.treevw.bind("<Alt-Down>",self.bDown)
        self.treevw.bind("<Alt-Up>",self.bUp)
        self.treevw.bind("<ButtonPress-1>",self.dragStart)
        self.treevw.bind("<ButtonRelease-1>",self.dragRelease, add='+')
        self.treevw.bind("<B1-Motion>",self.dragMove, add='+')
        self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
        scbVSel = CTkScrollbar(self.frameTw,
                                orientation=tk.VERTICAL,
                                command=self.treevw.yview)
        self.treevw.configure(yscrollcommand=scbVSel.set)
        scbVSel.grid(row=0, column=1, sticky=tk.NS)
        self.frameTw.pack(side=tk.TOP, fill=tk.BOTH, padx=5, pady=10)
        self.frameTw.columnconfigure(0, weight=1)
        self.frameTw.rowconfigure(0, weight=1)
        ### OFFICE EXPORT FRAME ###
        belowFrame = CTkFrame(self.paned)
        frameBtn = CTkFrame(self.frameTw)
        #lbl_help = FormHelper("DefectHelper", "Use del to delete a defect, use Alt+Arrows to order them")
        #lbl_help.constructView(frameBtn)
        self.buttonUpImage = CTkImage(Image.open(utils.getIcon("up-arrow.png")))
        self.buttonDownImage = CTkImage(Image.open(utils.getIcon('down-arrow.png')))
        # use self.buttonPhoto
        frame_up_down = CTkFrame(frameBtn)
        btn_down = CTkButton(frame_up_down,  text = "", width=20, image=self.buttonDownImage, command=self.bDown)
        btn_down.pack(side="left", anchor="center")
        btn_up = CTkButton(frame_up_down, text = "",  width=20, image=self.buttonUpImage, command=self.bUp)
        btn_up.pack(side="left", anchor="center")
        frame_up_down.pack(side=tk.TOP, anchor=tk.CENTER)
        btn_delDefect = CTkButton(
            frameBtn, text="Remove selection", command=self.deleteSelectedItem, image=self.image_del, fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        btn_delDefect.pack(side=tk.TOP, pady=5)
        btn_addDefect = CTkButton(
            frameBtn, text="Add defect", image=self.image_add, command=self.addDefectCallback)
        btn_addDefect.pack(side=tk.TOP, pady=5)
        btn_addDefect = CTkButton(
            frameBtn, text="Add Many defects", image=self.image_add, command=self.addManyDefectsCallback)
        btn_addDefect.pack(side=tk.TOP, pady=5)
        btn_setMainRedactor = CTkButton(
            frameBtn, text="Set main redactor", command=self.setMainRedactor, image=self.image_edit)
        btn_setMainRedactor.pack(side=tk.TOP, pady=5)
        btn_browseDefects = CTkButton(	
	            frameBtn, text="Browse defects templates", image=self.image_docs, command=self.browseDefectsCallback)	
        btn_browseDefects.pack(side=tk.TOP, pady=5)
        frameBtn.grid(row=0, column=2, sticky=tk.W + tk.E)
        officeFrame = ttk.LabelFrame(belowFrame, text=" Reports ")
        ### INFORMATION EXPORT FRAME ###
        informations_frame = CTkFrame(officeFrame)
        lbl_lang = CTkLabel(informations_frame, text="Lang")
        lbl_lang.grid(row=2, column=0, sticky=tk.E)
        self.combo_lang = CTkComboBox(informations_frame, values=self.langs, command=self.langChange)
        self.combo_lang.grid(row=2, column=1, sticky=tk.W)
        informations_frame.pack(side=tk.TOP, pady=10)
        ### WORD EXPORT FRAME ###
        templatesFrame = CTkFrame(officeFrame)
        lbl = CTkLabel(
            templatesFrame, text="Word template")
        lbl.grid(row=0, column=0, sticky=tk.E)
        self.combo_word = CTkComboBox(templatesFrame, values=self.docx_models, width=300)
        self.combo_word.grid(row=0, column=1)
        self.btn_template_photo = CTkImage(Image.open(utils.getIcon("download.png")))
        btn_word = CTkButton(
            templatesFrame, text="Generate", image=self.image_word, command=self.generateReportWord)
        btn_word.grid(row=0, column=2, sticky=tk.E, padx=5)
        btn_word_template_dl = CTkButton(templatesFrame, text="View empty template", width=40, image=self.btn_template_photo, command=self.downloadWordTemplate)
        btn_word_template_dl.grid(row=0, column=3, sticky=tk.W)
        
        ### POWERPOINT EXPORT FRAME ###
        lbl = CTkLabel(templatesFrame,
                        text="Powerpoint template")
        lbl.grid(row=1, column=0, sticky=tk.E, pady=20)
        self.combo_pptx = CTkComboBox(
            templatesFrame, values=self.pptx_models, width=300)
        self.combo_pptx.grid(row=1, column=1)
        
        btn_ppt = CTkButton(
            templatesFrame, text="Generate", image=self.image_ppt, command=self.generateReportPowerpoint)
        btn_ppt.grid(row=1, column=2, sticky=tk.E, padx=5)
        btn_pptx_template_dl = CTkButton(templatesFrame,text="View empty template",image=self.btn_template_photo,width=40, command=self.downloadPptxTemplate)
        btn_pptx_template_dl.grid(row=1, column=3, sticky=tk.W)
        templatesFrame.pack(side=tk.TOP,padx=10, pady=10, expand=1, anchor=tk.CENTER)
        officeFrame.pack(side=tk.TOP, fill=tk.BOTH, pady=10)
        belowFrame.pack(side=tk.TOP, fill=tk.BOTH)
        self.paned.add(self.frameTw, minsize=200)
        self.paned.add(belowFrame)
        self.paned.pack(fill=tk.BOTH, expand=1)
        defectLabelFrame.pack(side=tk.TOP, fill=tk.BOTH, pady=10)	
        frameAllBelow.pack(side=tk.TOP, fill=tk.BOTH)
        self.paned_remarks.add(self.remarkframeTw)	
        self.paned_remarks.add(frameAllBelow)	
        self.paned_remarks.pack(fill=tk.BOTH, expand=1)	
        self.remarksFrame.pack(side=tk.TOP, fill=tk.BOTH)
        self.reportFrame.pack(side=tk.TOP, fill=tk.BOTH, padx=10, pady=10, expand=1)
        

    def langChange(self, event=None):
        self.curr_lang = self.combo_lang.get()
        templates = APIClient.getInstance().getTemplateList(self.curr_lang)
        self.docx_models = [f for f in templates if f.endswith(".docx")]
        self.pptx_models = [f for f in templates if f.endswith(".pptx")]
        self.combo_word.configure(values=self.docx_models)
        self.combo_pptx.configure(values=self.pptx_models)
        self.settings.reloadSettings()
        pentest_type = self.settings.getPentestType().lower()
        
        pentesttype_docx_models = [f for f in self.docx_models if pentest_type in f.lower()]
        if pentesttype_docx_models:
            self.combo_word.set(pentesttype_docx_models[0])
        elif self.docx_models:
            self.combo_word.set(self.docx_models[0])

        pentesttype_pptx_models = [f for f in self.pptx_models if pentest_type in f.lower()]
        if pentesttype_pptx_models:
            self.combo_pptx.set(pentesttype_pptx_models[0])
        elif self.pptx_models:
            self.combo_pptx.set(self.pptx_models[0])

    def bDown(self, event=None):
        item_iid = self.treevw.selection()[0]
        children = self.treevw.get_children()
        iid_moving = children.index(item_iid)
        try:
            iid_moved_by = children[iid_moving+1]
            apiclient = APIClient.getInstance()
            apiclient.moveDefect(item_iid, iid_moved_by)
            self.treevw.move(item_iid, '', iid_moving+1)
        except IndexError:
            pass
        return "break"


    def bUp(self, event=None):
        item_iid = self.treevw.selection()[0]
        children = self.treevw.get_children()
        iid_moving = children.index(item_iid)
        try:
            iid_moved_by = children[iid_moving-1]
            apiclient = APIClient.getInstance()
            apiclient.moveDefect(item_iid, iid_moved_by)
            
            self.treevw.move(item_iid, '', iid_moving-1)
        except IndexError:
            pass
        return "break"

    def dragStart(self, event):
        tv = event.widget
        if tv.identify_row(event.y) not in tv.selection():
            tv.selection_set(tv.identify_row(event.y))    
            try:
                moving_item = tv.selection()[0]
            except IndexError:
                return # nothing selected
            moving_item = tv.item(moving_item)
            self.movingSelection = tv.identify_row(event.y)
            self.drag_toast = CTkLabel(self.parent, text=moving_item["text"])
            root = tv.winfo_toplevel()
            abs_coord_x = root.winfo_pointerx() - self.parent.winfo_rootx() + 10
            abs_coord_y = root.winfo_pointery() - self.parent.winfo_rooty() +3
            self.drag_toast.place(x=abs_coord_x,y=abs_coord_y)

    def dragRelease(self, event):
        if self.drag_toast is not None:
            self.drag_toast.destroy()
            self.drag_toast = None
        if self.movingSelection is None or self.lastMovedTo is None or self.lastMovedTo == "":
            return
        tv = event.widget
        apiclient = APIClient.getInstance()
        if tv.identify_row(event.y) in tv.get_children():
            apiclient.moveDefect(self.movingSelection, self.lastMovedTo)
            children = tv.get_children()
            iid_moving = children.index(self.lastMovedTo)
            tv.move(self.movingSelection, '', iid_moving)
            self.movingSelection = None
            self.lastMovedTo = None
            

    def dragMove(self, event):
        tv = event.widget
        rowToMove = tv.identify_row(event.y)
        moveto = tv.index(rowToMove)    
        self.lastMovedTo = rowToMove if rowToMove != self.movingSelection else self.lastMovedTo
        if self.drag_toast:
            root = tv.winfo_toplevel()
            abs_coord_x = root.winfo_pointerx() - self.parent.winfo_rootx() + 10
            abs_coord_y = root.winfo_pointery() - self.parent.winfo_rooty() +3
            self.drag_toast.place(x=abs_coord_x,y=abs_coord_y)
     

    def reset(self):
        """
        reset defect treeview by deleting every item inside.
        """
        for item in self.treevw.get_children():
            self.treevw.delete(item)

    def reset_remarks(self):	
        """	
        reset defect treeview by deleting every item inside.	
        """	
        for item in self.remarks_treevw.get_children():	
            self.remarks_treevw.delete(item)

    def deleteSelectedItem(self, _event=None):
        """
        Remove selected defect from treeview
        Args:
            _event: not used but mandatory
        """
        try:
            selected = self.treevw.selection()[0]
        except IndexError:
            return
        self.removeItem(selected)

    def deleteSelectedRemarkItem(self, _event=None):	
        """	
        Remove selected remark from treeview	
        Args:	
            _event: not used but mandatory	
        """	
        selected = self.remarks_treevw.selection()[0]	
        self.remarks_treevw.delete(selected)	
        remark = Remark.fetchObject({"_id":ObjectId(selected)})	
        remark.delete()	
        self.resizeDefectTreeview()

    def removeItem(self, toDeleteIid):
        """
        Remove defect from given iid in defect treeview
        Args:
            toDeleteIid: database ID of defect to delete
        """
        try:
            item = self.treevw.item(toDeleteIid)
        except tk.TclError:
            return
        dialog = ChildDialogQuestion(self.parent,
                                     "DELETE WARNING", "Are you sure you want to delete defect "+str(item["text"])+" ?", ["Delete", "Cancel"])
        self.parent.wait_window(dialog.app)
        if dialog.rvalue != "Delete":
            return
        self.treevw.delete(toDeleteIid)
        defectToDelete = Defect.fetchObject({"title": item["text"], "target_id":""})
        if defectToDelete is not None:
            defectToDelete.delete()
            self.resizeDefectTreeview()

    def removeRemarkItem(self, toDeleteIid):	
        """	
        Remove remark from given iid in defect treeview	
        Args:	
            toDeleteIid: database ID of defect to delete	
        """	
        item = self.remarks_treevw.item(toDeleteIid)	
        dialog = ChildDialogQuestion(self.parent,	
                                        "DELETE WARNING", "Are you sure you want to delete remark "+str(item["text"])+" ?", ["Delete", "Cancel"])	
        self.parent.wait_window(dialog.app)	
        if dialog.rvalue != "Delete":	
            return	
        self.remarks_treevw.delete(toDeleteIid)	
        remarkToDelete = Remark.fetchObject({"title": item["text"]})	
        if remarkToDelete is not None:	
            remarkToDelete.delete()	
            self.resizeRemarkTreeview()

    def addDefectCallback(self):
        """Open an insert defect view form in a child window"""
        dialog = ChildDialogDefectView(self.tkApp, "Add a security defect", self.settings)
        try:
            self.tkApp.wait_window(dialog.app)	
        except tk.TclError:
            pass

    def browseDefectsCallback(self):	
        """Open an view to browse defects Templates"""	
        dialog = ChildDialogDefectView(self.tkApp, "Browse defect templates", self.settings, Defect(), True)	
        try:
            self.tkApp.wait_window(dialog.app)	
        except tk.TclError:
            pass

    def addManyDefectsCallback(self):
        """Open an multiview insert defect view form in a child window"""	
        dialog = ChildDialogDefectView(self.tkApp, "Add and edit multiple defect", self.settings, None, True)	
        try:
            self.tkApp.wait_window(dialog.app)	
        except tk.TclError:
            pass
		
    def addRemarkCallback(self):	
        """Open an insert defect view form in a child window"""	
        dialog = ChildDialogRemarkView(self.parent)

    def setMainRedactor(self):
        """Sets a main redactor for a pentest. Each not assigned defect will be assigned to him/her"""
        self.settings.reloadSettings()
        dialog = ChildDialogCombo(self.parent, self.settings.getPentesters()+["N/A"], "Set main redactor", "N/A")
        self.parent.wait_window(dialog.app)
        newVal = dialog.rvalue
        if newVal is None:
            return
        if not newVal or newVal.strip() == "":
            return
        columnRedactor = self.treevw['columns'].index("redactor")
        for it in self.treevw.get_children():
            oldValues = self.treevw.item(it)["values"]
            if oldValues[columnRedactor] == "N/A":
                oldValues[columnRedactor] = newVal
                self.treevw.item(it, values=oldValues)
                d_o = Defect({"_id":it})
                d_o.update({"redactor":newVal})
        self.mainRedac = newVal

    def updateDefectInTreevw(self, defect_m, redactor=None):
        """
        Change values of a selected defect in the treeview
        Args:
            defect_m: a defect model with updated values
            redactor: a redactor name for this defect, can be None (default)
        """
        if defect_m.isAssigned():
            return ""
        try:
            exist = self.treevw.item(defect_m.getId())
        except:
            self.addDefect(defect_m)
            return
        try:
            columnEase = self.treevw['columns'].index("ease")
            columnImpact = self.treevw['columns'].index("impact")
            columnRisk = self.treevw['columns'].index("risk")
            columnType = self.treevw['columns'].index("type")
            columnRedactor = self.treevw['columns'].index("redactor")
        except :
            return
        oldValues = self.treevw.item(defect_m.getId())["values"]
        oldRisk = oldValues[columnRisk]
        newRisk = defect_m.risk
        newValues = [""]*5
        newValues[columnEase] = defect_m.ease
        newValues[columnImpact] = defect_m.impact
        newValues[columnRisk] = defect_m.risk
        newValues[columnType] = ", ".join(defect_m.mtype)
        newValues[columnRedactor] = defect_m.redactor
        self.treevw.item(defect_m.getId(), text=defect_m.title, tags=(newRisk), values=newValues)
        #if self.movingSelection is None:
        self.treevw.move(defect_m.getId(), '', int(defect_m.index))

    def updateRemarkInTreevw(self, remark_m):
        """
        Change values of a selected remark in the treeview
        Args:
            remark_m: a remark model with updated values
        """
        try:
            exist = self.remarks_treevw.item(remark_m.getId())
        except:
            self.addRemark(remark_m)
            return
        self.remarks_treevw.item(remark_m.getId(), text=remark_m.title, values=[remark_m.type], image=RemarkView.getIcon(remark_m.type))

    def OnRemarkDoubleClick(self, event):
        """Callback for double click on remark treeview. Opens a window to update the double clicked remark view.
        """
        item = self.remarks_treevw.identify("item", event.x, event.y)
        if item is None or item == '':
            return
        remark_m = Remark.fetchObject({"_id": ObjectId(item)})
        dialog = ChildDialogRemarkView(self.tkApp,  remark_m)
        self.parent.wait_window(dialog.app)
        self.updateRemarkInTreevw(remark_m)
      
    def OnDoubleClick(self, event):
        """
        Callback for double click on treeview.
        Opens a window to update the double clicked defect view.
        Args:
            event: automatically created with the event catch. stores data about line in treeview that was double clicked.
        """
        item = self.treevw.identify("item", event.x, event.y)
        if item is None or item == '':
            return
        defect_m = Defect.fetchObject({"_id": ObjectId(item)})
        dialog = ChildDialogDefectView(self.tkApp, "Edit defect", self.settings, defect_m)
        self.updateDefectInTreevw(defect_m)

    def fillWithDefects(self):
        """
        Fetch defects that are global (not assigned to an ip) and fill the defect table with them.
        """
        table = Defect.getDefectTable()
        for line in table:
            self.addDefect(Defect(line))

    def fillWithRemarks(self):	
        """	
        Fetch remarks and fill the remarks table with them.	
        """	
        self.reset_remarks()
        remarks = Remark.fetchObjects({})	
        for remark in remarks:	
            self.addRemark(remark)
    
    def addRemark(self, remark_o):	
        type_of_remark = remark_o.type	
        already_inserted = False	
        already_inserted_iid = None	
        children = self.remarks_treevw.get_children()	
        for child in children:	
            title = self.remarks_treevw.item(child)["text"]	
            if title == remark_o.title:	
                already_inserted = True
                break	
        if not already_inserted:	
            try:	
                self.remarks_treevw.insert('', 'end', str(remark_o.getId()), values = (remark_o.type), text=remark_o.title, image=RemarkView.getIcon(remark_o.type))	
            except tk.TclError:	
                # The defect already exists	
                self.remarks_treevw.item(str(remark_o.getId()), values = (remark_o.type), text=remark_o.title, image=RemarkView.getIcon(remark_o.type))
            
        self.resizeRemarkTreeview()	
                    
	
	
    def addDefect(self, defect_o):
        """
        Add the given defect object in the treeview
        Args:
            defect_o: a Models.Defect object to be inserted in treeview
        """
        if defect_o is None:
            return
        if not self.inited:
            return
        children = self.treevw.get_children()
        indToInsert = defect_o.index
        types = defect_o.mtype
        types = ", ".join(defect_o.mtype)
        new_values = (defect_o.ease, defect_o.impact,
                      defect_o.risk, types, defect_o.redactor if defect_o.redactor != "N/A" else self.mainRedac)
        already_inserted = False
        already_inserted_iid = None
        for child in children:
            title = self.treevw.item(child)["text"]
            if title == defect_o.title:
                already_inserted = True
                already_inserted_iid = child
                break
        if not already_inserted:
            try:
                self.treevw.insert('', int(indToInsert), defect_o.getId(), text=defect_o.title,
                                   values=new_values,
                                   tags=(defect_o.risk))
            except tk.TclError:
                # The defect already exists
                already_inserted = True
                already_inserted_iid = defect_o.getId()
        if already_inserted:
            existing = self.treevw.item(already_inserted_iid)
            values = existing["values"]
            if values[4].strip() == "N/A":
                values[4] = defect_o.redactor
            elif defect_o.redactor not in values[4].split(", "):
                values[4] += ", "+defect_o.redactor
            self.treevw.item(already_inserted_iid, values=values)
        self.resizeDefectTreeview()
    
    def resizeDefectTreeview(self):
        currentHeight = len(self.treevw.get_children())
        if currentHeight <= 15:
            self.treevw.configure(height=currentHeight)
            sx, sy = self.paned.sash_coord(0)
            if sy <= (currentHeight)*self.rowHeight + self.pane_base_height:
                self.paned.paneconfigure(self.frameTw, height=(currentHeight)*self.rowHeight + self.pane_base_height)
    
    def resizeRemarkTreeview(self):	
        currentHeight = len(self.remarks_treevw.get_children())	
        if currentHeight <= 5:	
            self.remarks_treevw.configure(height=currentHeight)	
            sx, sy = self.paned_remarks.sash_coord(0)	
            if sy <= (currentHeight)*self.rowHeight + self.pane_base_height:	
                self.paned_remarks.paneconfigure(self.remarkframeTw, height=(currentHeight)*self.rowHeight + self.pane_base_height)

    def generateReportPowerpoint(self):
        apiclient = APIClient.getInstance()
        toExport = apiclient.getCurrentPentest()
        if toExport != "":
            modele_pptx = str(self.combo_pptx.get())
            dialog = ChildDialogInfo(
                self.parent, "PowerPoint Report", "Creating report . Please wait.")
            dialog.show()
            x = threading.Thread(target=generateReport, args=(self.tkApp, dialog, modele_pptx, self.mainRedac, self.curr_lang))
            x.start()

    def generateReportWord(self):
        """
        Export a pentest defects to a word formatted file.
        """
        
        apiclient = APIClient.getInstance()
        toExport = apiclient.getCurrentPentest()
        if toExport != "":
            modele_docx = str(self.combo_word.get())
            dialog = ChildDialogInfo(
                self.parent, "Word Report", "Creating report . Please wait.")
            dialog.show()
            x = threading.Thread(target=generateReport, args=(self.tkApp, dialog, modele_docx,  self.mainRedac, self.curr_lang))
            x.start()
            

    def downloadWordTemplate(self):
        self._download_and_open_template(self.combo_word.get())

    def downloadPptxTemplate(self):
        self._download_and_open_template(self.combo_pptx.get())

    def _download_and_open_template(self, templateName):
        apiclient = APIClient.getInstance()
        path = apiclient.downloadTemplate(self, self.curr_lang, templateName)
        dialog = ChildDialogQuestion(self.parent,
                                    "Template downloaded", "Template was downloaded here : "+str(path)+". Do you you want to open it ?", ["Open", "Cancel"])
        self.parent.wait_window(dialog.app)
        if dialog.rvalue != "Open":
            return
        utils.openPathForUser(path, folder_only=True)

    def update_received(self, dataManager, notif, obj, old_obj):
        if notif["collection"] == "remarks":
            obj = Remark.fetchObject({"_id":notif["iid"]})
        if obj is None:
            return
        
        if notif["action"] == "delete":
            if notif["collection"] == "remarks":
                self.deleteRemarkInTreevw(obj)
            elif notif["collection"] == "defects":
                self.deleteDefectInTreevw(obj)
            return
        # insert ou update
        if notif["collection"] == "remarks":
            logger.debug("Received remark notification : "+str(notif))
            self.addRemark(obj)
        elif notif["collection"] == "defects":
            self.updateDefectInTreevw(obj)
      
        


def generateReport(tkApp, dialog, modele, mainRedac, curr_lang):
    apiclient = APIClient.getInstance()
    settings = Settings()
    settings._reloadDbSettings()
    additional_context = {}
    for module in tkApp.modules:
        if callable(getattr(module["object"], "onGenerateReport", None)):
                result = module["object"].onGenerateReport()
                if result.get("additional_context") is not None:
                    additional_context = {**result.get("additional_context", {}), **additional_context}
    res, msg = apiclient.generateReport(modele, settings.getClientName(), settings.getMissionName(), mainRedac, curr_lang, additional_context)
    dialog.destroy()
    if not res:
        tkinter.messagebox.showerror(
            "Failure", str(msg))
    else:
        tkinter.messagebox.showinfo(
            "Success", "The document was generated in "+str(msg))
        utils.openPathForUser(msg, folder_only=True)
        


"""Controller for IP object. Mostly handles conversion between mongo data and python objects"""

from pollenisatorgui.core.controllers.controllerelement import ControllerElement
from pollenisatorgui.core.models.ip import Ip
import json

class IpController(ControllerElement):
    """Inherits ControllerElement
    Controller for IP object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Ip represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by IpView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Ip document.
        """
        self.model.notes = values.get("Notes", self.model.notes)
        self.model.infos = values.get("Infos", self.model.infos)
        if not isinstance(self.model.infos, dict):
            self.model.infos = json.loads(self.model.infos)
        for info in self.model.infos:
            self.model.infos[info] = self.model.infos[info][0]
        self.model.update()

    def doInsert(self, values):
        """
        Insert the Ip represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by MultipleIpView or IpView containg all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        # Only multi insert exists at the moment for IP
        try:
            multi = True
        except KeyError:
            multi = False

        if multi:
            # Get form values
            ret = []
            total = 0
            accepted = 0
            for line in values["IPs"].split("\n"):
                if line != "":
                    # Insert in database
                    model = Ip().initialize(line)
                    inserted, iid = model.addInDb()
                    if inserted:
                        ret.append(iid)
                        accepted += 1
                    total += 1
            return ret, total-accepted  # nb errors = total - accepted

    def is_in_scope(self):
        """Return true if the ip has registered at least one scope in its field 'in_scopes' , False otherwise"""
        return bool(self.model.in_scopes)
    
    def getDefects(self):
        """Return ip assigned defects as a list of mongo fetched defects dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return self.model.getDefects()

    def getChecks(self):
        """Return ip assigned checks as a list of mongo fetched checks instance
        Returns:
            list of checkInstance objects
        """
        return self.model.getChecks()

    def getPorts(self):
        """Return ip assigned ports as a list of mongo fetched defects dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return self.model.getPorts()

    def getType(self):
        """Return a string describing the type of object
        Returns:
            "ip" """
        return "ip"


from pollenisatorgui.core.plugins.plugin import Plugin
from pollenisatorgui.core.models.port import Port
import webbrowser

class Dirsearch(Plugin):

    def __init__(self):
        """Constructor"""
        self.toolmodel = None
        self.port_m = None

    def getActions(self, toolmodel):
        """
        Summary: Add buttons to the tool view.
        Args:
            toolmodel : the tool model opened in the pollenisator client.
        Return:
            A dictionary with buttons text as key and function callback as value.
        """
        self.toolmodel = toolmodel
        self.port_m = Port.fetchObject(
            {"ip": toolmodel.ip, "port": toolmodel.port, "proto": toolmodel.proto})
        if self.port_m is None:
            return {}
        return {"Open 200 in browser": self.openInBrowser}

    def openInBrowser(self, _event=None):
        """Callback of action  Open 200 in browser
        Open scanned host port in browser as tabs.
        Args:
            _event: not used but mandatory
        """
        if self.port_m is None:
            return
        ssl = self.port_m.infos.get("SSL", None) == "True" or ("https" in self.port_m.service or "ssl" in self.port_m.service)
        protocol = "https://" if ssl else "http://"
        dirs = self.port_m.infos.get("Dirsearch_200", [])
        if len(dirs) > 10:
            return "Too much 200 to be opened this way"
        for finding in dirs:
            url = protocol+self.port_m.ip+":"+str(self.port_m.port)+finding
            webbrowser.open_new_tab(url)


from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.datamanager import DataManager

REGISTRY = {}

def register_class(target_class):
    """Register the given class
    Args:
        target_class: type <class>
    """
    REGISTRY[target_class.__name__] = target_class

class MetaModule(type):
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        if name not in REGISTRY:
            register_class(cls)
        return cls

class Module(metaclass=MetaModule):
    need_admin = False
    pentest_types = ["all"]
    FIRST_PRIORITY = 0
    HIGH_PRIORITY = 1
    MEDIUM_PRIORITY = 2
    LOW_PRIORITY = 3
    LAST_PRIORITY = 99
    order_priority = LOW_PRIORITY
    def __init__(self):
        datamanager = DataManager.getInstance().attach(self)
        
    def update_received(self, dataManager, notif, obj, old_obj):
        pass
    



"""Wave Model. Stores which command should be launched and associates Interval and Scope"""

from bson.objectid import ObjectId
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.models.ip import Ip
from pollenisatorgui.core.models.interval import Interval
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.models.scope import Scope
from pollenisatorgui.core.components.apiclient import APIClient


class Wave(Element):
    """
    Represents a Wave object. A wave is a series of tools to execute.

    Attributes:
        coll_name: collection name in pollenisator database
    """
    coll_name = "waves"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched interval is optimal.
                        possible keys with default values are : _id(None), parent(None), tags([]), infos({}),
                        wave(""), wave_commands([])
        """
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None), valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("wave", ""),
                        valuesFromDb.get("wave_commands", []), valuesFromDb.get("infos", {}))

    def initialize(self, wave="", wave_commands=None, infos=None):
        """Set values of scope
        Args:
            wave: the wave name, default is ""
            wave_commands: a list of command name that are to be launched in this wave. Defaut is None (empty list)
            infos: a dictionnary of additional info. Default is None (empty dict)
        Returns:
            this object
        """
        self.wave = wave
        self.wave_commands = wave_commands if wave_commands is not None else []
        self.infos = infos if infos is not None else {}
        return self

    def getData(self):
        return{"wave": self.wave, "wave_commands": self.wave_commands, "_id": self.getId(), "infos": self.infos}


    def delete(self):
        """
        Delete the wave represented by this model in database.
        Also delete the tools, intervals, scopes associated with this wave
        """
        apiclient = APIClient.getInstance()
        apiclient.delete("waves", ObjectId(self._id))

    def addInDb(self):
        """
        Add this wave in database.
        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        apiclient = APIClient.getInstance()
        res, iid = apiclient.insert(
            "waves", {"wave": self.wave, "wave_commands": list(self.wave_commands)})
        self._id = iid
        return res, iid

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            apiclient.update("waves", ObjectId(self._id), {"wave_commands": list(self.wave_commands)})
        else:
            apiclient.update("waves", ObjectId(self._id), pipeline_set)


    def __str__(self):
        """
        Get a string representation of a wave.

        Returns:
            Returns the wave id (name).
        """
        return self.wave

 

    def getAllTools(self):
        """Return all tools being part of this wave as a list of mongo fetched tools dict.
        Differs from getTools as it fetches all tools of the name and not only tools of level wave.
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return Tool.fetchObjects({"wave": self.wave})

    def getIntervals(self):
        """Return scope assigned intervals as a list of mongo fetched intervals dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        datamanager = DataManager.getInstance()
        return datamanager.find("intervals",
                                  {"wave": self.wave})

    def getScopes(self):
        """Return wave assigned scopes as a list of mongo fetched scopes dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        datamanager = DataManager.getInstance()
        return datamanager.find("scopes", {"wave": self.wave})

    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (1 key :"wave")
        """
        return {"wave": self.wave}

    def isLaunchableNow(self):
        """Returns True if the tool matches criteria to be launched 
        (current time matches one of interval object assigned to this wave)
        Returns:
            bool
        """
        intervals = Interval.fetchObjects({"wave": self.wave})
        for intervalModel in intervals:
            if utils.fitNowTime(intervalModel.dated, intervalModel.datef):
                return True
        return False



    @classmethod
    def listWaves(cls):
        """Return all waves names as a list 
        Returns:
            list of all wave names
        """
        ret = []
        datamanager = DataManager.getInstance()
        waves = datamanager.get("waves", '*')
        for wave in waves:
            ret.append(wave["wave"])
        return ret

  



import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from PIL import Image
from pollenisatorgui.core.application.dialogs.ChildDialogSelectChecks import ChildDialogSelectChecks
from pollenisatorgui.core.components import utils
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.tag import TagInfos
from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.modules.module import Module
from pollenisatorgui.core.forms.formpanel import FormPanel
import threading


class Dashboard(Module):
    """
    Shows information about ongoing pentest. 
    """
    iconName = "tab_dashboard.png"
    tabName = "Dashboard"
    order_priority = Module.FIRST_PRIORITY

    def __init__(self, parent, settings, tkApp):
        """
        Constructor
        """
        super().__init__()
        self.timer = None
        self.mainApp = parent
        self.parent = None
        self.parent = parent
        
        self.tkApp = tkApp
        self.treevwApp = None
        self.autoscan_slider = None
        self.settings = settings
        self.label_count_vuln = None
        self.inited = False

    def open(self, view, topviewframe, treevw):
        apiclient = APIClient.getInstance()
        self.treevwApp = treevw
        if self.inited is False:
            self.initUI(view)
        if apiclient.getCurrentPentest() is not None:
            self.refreshUI()

        return True

    def close(self):
        if self.timer is not None:
            self.timer.cancel()
            self.timer = None

    def refreshUI(self):
        """
        Reload data and display them
        """
        if self.tkApp.quitting:
            return
        self.loadData()
        self.displayData()
        #self.timer = threading.Timer(3.0, self.refreshUI)
        # self.timer.start()

    def loadData(self):
        """
        Fetch data from database
        """

        self.infos = APIClient.getInstance().getGeneralInformation()

    def displayData(self):
        """
        Display loaded data in treeviews
        """
        self.set_vuln_count()
        self.set_autoscan_status()
        self.set_scan_progression()
        self.set_cheatsheet_progression()
        self.set_results()

    
    def initUI(self, parent):
        """
        Build UI
        Args:
            parent: its parent widget
        """
        self.inited = True
        self.moduleFrame = CTkFrame(parent)
        vuln_frame = CTkFrame(self.moduleFrame, height=0)
        self.populate_vuln_frame(vuln_frame)
        vuln_frame.pack(side=tk.TOP, anchor=tk.W, ipady=10)
        autoscan_frame = CTkFrame(self.moduleFrame, height=0)
        self.populate_autoscan_frame(autoscan_frame)
        autoscan_frame.pack(side=tk.TOP, ipady=10, pady=10)
        results_frame = CTkFrame(self.moduleFrame, height=0)
        self.populate_results_frame(results_frame)
        results_frame.pack(side=tk.TOP, ipady=10, pady=10,
                           fill=tk.BOTH, expand=True)
        self.moduleFrame.pack(padx=10, pady=10, side="top",
                              fill=tk.BOTH, expand=True)

    def populate_vuln_frame(self, vuln_frame):
        self.vuln_image = CTkImage(Image.open(utils.getIcon("defect.png")))
        self.host_image = CTkImage(Image.open(utils.getIcon("ip.png")))
        self.label_host_count = CTkLabel(
            vuln_frame, text="X Hosts", image=self.host_image, compound="left")
        self.label_host_count.pack(padx=10, pady=3, side=tk.TOP, anchor=tk.W)
        self.label_count_vuln = CTkLabel(
            vuln_frame, image=self.vuln_image, compound="left", text="X Vulnerabilities")
        self.label_count_vuln.pack(padx=10, pady=3, side=tk.TOP, anchor=tk.W)
        sub_frame = CTkFrame(vuln_frame)
        self.label_count_vuln_critical = CTkLabel(
            sub_frame, text="X Critical", fg_color="black", text_color="white")
        self.label_count_vuln_critical.cget("font").configure(weight="bold")
        self.label_count_vuln_major = CTkLabel(
            sub_frame, text="X Major", fg_color="red", text_color="white")
        self.label_count_vuln_major.cget("font").configure(weight="bold")
        self.label_count_vuln_important = CTkLabel(
            sub_frame, text="X Important", fg_color="orange", text_color="white")
        self.label_count_vuln_important.cget("font").configure(weight="bold")
        self.label_count_vuln_minor = CTkLabel(
            sub_frame, text="X Minor", fg_color="yellow", text_color="black")
        self.label_count_vuln_minor.cget("font").configure(weight="bold")
        self.label_count_vuln_critical.pack(side="left", padx=3, ipadx=3)
        self.label_count_vuln_major.pack(side="left", padx=3, ipadx=3)
        self.label_count_vuln_important.pack(side="left", padx=3, ipadx=3)
        self.label_count_vuln_minor.pack(side="left", padx=3, ipadx=3)
        sub_frame.pack(padx=3, pady=1, side=tk.TOP, anchor=tk.W)

    def set_vuln_count(self):
        try:
            self.label_count_vuln.configure(
                text=str(self.infos.get("defect_count", 0))+" Vulnerabilities")
            self.label_count_vuln_critical.configure(
                text=str(self.infos.get("defect_count_critical", 0))+" Critical")
            self.label_count_vuln_major.configure(
                text=str(self.infos.get("defect_count_major", 0))+" Major")
            self.label_count_vuln_important.configure(
                text=str(self.infos.get("defect_count_important", 0))+" Important")
            self.label_count_vuln_minor.configure(
                text=str(self.infos.get("defect_count_minor", 0))+" Minor")
        except tk.TclError:
            return

    def slider_event(self, event):
        val = self.autoscan_slider.get()
        val = int(val)
        self.autoscan_threads_lbl.configure(
            text=str(val) + " thread"+("s" if val > 1 else ""))
        self.settings.setPentestSetting("autoscan_threads", val)

    def update_thread_label(self, event):
        val = self.autoscan_slider.get()
        # behavior is weird .set(4) then .get() return 3.7
        val = int(round(val))
        self.autoscan_threads_lbl.configure(
            text=str(val) + " thread"+("s" if val > 1 else ""))

    def populate_autoscan_frame(self, frame):
        self.image_auto = CTkImage(Image.open(utils.getIcon("auto.png")))
        self.image_start = CTkImage(Image.open(utils.getIcon("start.png")))
        self.image_stop = CTkImage(Image.open(utils.getIcon("stop.png")))
        self.image_pentest = CTkImage(Image.open(utils.getIcon("hacker.png")))
        frame_status = FormPanel(side=tk.TOP, fill=tk.X)
        frame_status.addFormButton("Go to pentest", self.go_to_pentest, image=self.image_pentest, width=200, height=50, padx=5, side=tk.LEFT, font=CTkFont("Calibri", 20, 'bold', "roman", True),anchor="center")
        frame_status.addFormSeparator(orient="vertical", padx=0, fill=tk.Y, side=tk.LEFT)
        autoscan_panel = frame_status.addFormPanel(side=tk.LEFT, fill=tk.X, padx=0, pady=0, anchor="center")
        autoscan_action_panel = autoscan_panel.addFormPanel(side=tk.TOP, fill=tk.X, padx=0, pady=0, anchor="n")
        autoscan_action_panel.addFormLabel("Autoscan", image=self.image_auto, compound="left" , padx=5, side=tk.LEFT )
        self.btn_autoscan = autoscan_action_panel.addFormButton("autoscan", self.set_autoscan_status, text="", image=self.image_auto, padx=0, side=tk.LEFT)
        self.set_autoscan_status()
        autoscan_action_panel.addFormHelper("Autoscan will launch queued scans on available workers automatically. Queueing scan must be done manually.", padx=0, pady=0, side=tk.RIGHT)
        frame_status.constructView(frame)
        frame_settings = CTkFrame(autoscan_panel.panel)
        self.autoscan_slider = CTkSlider(
            frame_settings, from_=1, to=10, number_of_steps=10, command=self.update_thread_label)
        self.autoscan_slider.bind("<ButtonRelease-1>", self.slider_event)
        self.autoscan_slider.pack(side=tk.LEFT, padx=5)
        threads = int(self.settings.db_settings.get("autoscan_threads", 4))
        self.autoscan_slider.set(threads)
        self.autoscan_threads_lbl = CTkLabel(frame_settings, text=str(
            threads)+" thread"+("s" if threads > 1 else ""))
        self.autoscan_threads_lbl.pack(side=tk.LEFT, padx=5)
        frame_settings.pack(side=tk.BOTTOM, fill=tk.X)
        
        
        frame_progress = CTkFrame(frame)
        lbl = CTkLabel(frame_progress, text="Scan Progression")
        lbl.pack(side=tk.LEFT, padx=5)
        self.scan_progressbar = CTkProgressBar(
            frame_progress, mode='determinate')
        self.scan_progressbar.pack(
            side=tk.LEFT, padx=5, fill=tk.X, expand=True)
        self.label_scan_progress = CTkLabel(frame_progress, text="X/X")
        self.label_scan_progress.pack(side=tk.LEFT, padx=2)
        frame_progress.pack(side=tk.TOP, pady=5)
        frame_progress_cheatsheet = CTkFrame(frame)
        lbl = CTkLabel(frame_progress_cheatsheet,
                       text="Cheatsheet Progression")
        lbl.pack(side=tk.LEFT, padx=5)
        self.cheatsheet_progressbar = CTkProgressBar(
            frame_progress_cheatsheet, mode='determinate')
        self.cheatsheet_progressbar.pack(
            side=tk.LEFT, padx=5, fill=tk.X, expand=True)
        self.label_cheatsheet_progress = CTkLabel(
            frame_progress_cheatsheet, text="X/X")
        self.label_cheatsheet_progress.pack(side=tk.LEFT, padx=2)
        frame_progress_cheatsheet.pack(side=tk.TOP, pady=5)

    def set_autoscan_status(self, event=None):
        apiclient = APIClient.getInstance()
        status = apiclient.getAutoScanStatus()
        if status:
            self.btn_autoscan.configure(
                text="Stop autoscan", command=self.click_stop_autoscan, image=self.image_stop)
        else:
            self.btn_autoscan.configure(
                text="Start autoscan", command=self.click_start_autoscan, image=self.image_start)
        if self.autoscan_slider:
            value = int(self.settings.db_settings.get("autoscan_threads", 4))
            self.autoscan_slider.set(value)
            self.update_thread_label(value)

    def click_stop_autoscan(self):
        res = self.tkApp.stop_autoscan()
        if res:
            self.set_autoscan_status()

    def click_start_autoscan(self):

        res = self.tkApp.start_autoscan()
        if res:
            self.set_autoscan_status()

    def go_to_pentest(self, event=None):
        self.tkApp.nbk.select("Main View")


    def set_scan_progression(self):
        done = self.infos.get("tools_done_count", 0)
        total = float(self.infos.get("tools_count", 0))
        if total > 0:
            self.scan_progressbar.set(float(done)/float(total))
        self.scan_progressbar.update_idletasks()
        self.label_scan_progress.configure(text=str(done)+"/"+str(int(total)))
        self.label_scan_progress.update_idletasks()

    def set_cheatsheet_progression(self):
        done = self.infos.get("checks_done", 0)
        total = float(self.infos.get("checks_total", 0))
        if total > 0:
            self.cheatsheet_progressbar.set(float(done)/float(total))
        self.cheatsheet_progressbar.update_idletasks()
        self.label_cheatsheet_progress.configure(
            text=str(done)+"/"+str(int(total)))
        self.label_cheatsheet_progress.update_idletasks()

    def populate_results_frame(self, frame):
        self.form = FormPanel(fill="x",pady=0)
        form_top = self.form.addFormPanel(grid=True, side="top", fill="x", pady=0, anchor="s")
        form_top.addFormLabel("Text filter", row=0, column=0, sticky="s")
        self.str_filter = form_top.addFormStr("text_filter", placeholder="high", binds={"<KeyRelease>":self.filter}, width=300, row=0, column=1, sticky="S")
        form_bottom = self.form.addFormPanel(side="bottom", fill="x", pady=0)
        values = ["Critical", "Major", "Important", "Minor"]
        tags_registered = Settings.getTags()
        for tag_info in tags_registered.values():
            if tag_info["level"] not in values and tag_info["level"] != "":
                values.append(tag_info["level"])
        check_panel = form_top.addFormPanel(row=0, column=2,fill=None)
        self.severity_btns = check_panel.addFormChecklist("Severity", list(values), command=self.filter)

        self.treeview = form_bottom.addFormTreevw(
            "results", ("Result type", "Time", "Severity", "Title", "Target"), binds={"<Double-Button-1>": self.on_result_double_click},status="readonly", height=10, fill="x", side="top", expand=True)  
        self.form.constructView(frame)
        
    def filter(self, event=None):
        str_filter = self.str_filter.getValue()
        risks_checked = [k for k,v in self.severity_btns.getValue().items() if v == 1]
        self.treeview.filter(str_filter, str_filter, str_filter, str_filter, str_filter, check_all=False)
        if len(risks_checked) > 0:
            self.treeview.filter(True, True, risks_checked, True, True, reset=False)

    def on_result_double_click(self, event):
        if self.treeview is not None:
            tv = event.widget
            item = tv.identify("item", event.x, event.y)
            if tv.item(item)["text"] == "Tag":
                tag = self.datamanager.get("tags", item.split("|")[0])
                id_to_show = tag.item_id
            else:
                id_to_show = item
            self.mainApp.search("id == \""+str(id_to_show)+"\"")

    def set_results(self):
        self.datamanager = DataManager.getInstance()
        self.label_host_count.configure(
            text="Hosts : "+str(self.infos.get("hosts_count", 0)))
        defects = self.datamanager.get("defects", '*')
        self.treeview.reset()
        for defect in defects:
            if defect.isAssigned():
                self.treeview.addItem("", tk.END, defect.getId(), text="Security Defect", values=(
                    defect.creation_time, defect.risk, defect.title, defect.getDetailedString(onlyTarget=True)))
        tags = self.infos.get("tagged", [])
        tags_registered = Settings.getTags()
        for tag_infos in tags:
            for tag in tag_infos.get("tags", []):
                tag = TagInfos(tag)
                tag_name = tag.name
                
                if tags_registered.get(tag_name, {}).get("level", "") != "":
                    self.treeview.addItem("", tk.END, tag_infos["_id"]+"|"+tag_name, text="Tag", values=(
                        tag_infos["date"], tags_registered[tag_name]["level"], tag_name,  tag_infos["detailed_string"]))
        self.treeview.auto_resize_columns()